import logging
import multiprocessing as mp
import time
from typing import Literal, Optional, Tuple, Type, Union

import numpy as np
import numpy.typing as npt
import pyarrow.compute as pc
import quivr as qv
import ray
from scipy.linalg import solve

from ..coordinates import CartesianCoordinates, CoordinateCovariances
from ..coordinates.residuals import Residuals
from ..orbit_determination import OrbitDeterminationObservations
from ..orbits import Orbits
from ..propagator import Propagator
from ..ray_cluster import initialize_use_ray
from ..utils.iter import _iterate_chunk_indices, _iterate_chunks
from .fitted_orbits import FittedOrbitMembers, FittedOrbits
from .outliers import calculate_max_outliers

logger = logging.getLogger(__name__)

__all__ = ["differential_correction"]


def od_worker(
    orbit_ids: npt.NDArray[np.str_],
    orbits: FittedOrbits,
    orbit_members: FittedOrbitMembers,
    observations: OrbitDeterminationObservations,
    propagator: Type[Propagator],
    rchi2_threshold: float = 100,
    min_obs: int = 5,
    min_arc_length: float = 1.0,
    contamination_percentage: float = 0.0,
    delta: float = 1e-6,
    max_iter: int = 20,
    method: Literal["central", "finite"] = "central",
    propagator_kwargs: dict = {},
) -> Tuple[FittedOrbits, FittedOrbitMembers]:

    od_orbits = FittedOrbits.empty()
    od_orbit_members = FittedOrbitMembers.empty()
    for orbit_id in orbit_ids:
        time_start = time.time()
        logger.debug(f"Differentially correcting orbit {orbit_id}...")

        orbit = orbits.select("orbit_id", orbit_id)
        obs_ids = orbit_members.apply_mask(
            pc.equal(orbit_members.orbit_id, orbit_id)
        ).obs_id
        orbit_observations = observations.apply_mask(pc.is_in(observations.id, obs_ids))

        # Sort observations by time
        orbit_observations = orbit_observations.sort_by(
            [
                "coordinates.time.days",
                "coordinates.time.nanos",
                "coordinates.origin.code",
            ]
        )

        od_orbit, od_orbit_orbit_members = od(
            orbit,
            orbit_observations,
            rchi2_threshold=rchi2_threshold,
            min_obs=min_obs,
            min_arc_length=min_arc_length,
            contamination_percentage=contamination_percentage,
            delta=delta,
            max_iter=max_iter,
            method=method,
            propagator=propagator,
            propagator_kwargs=propagator_kwargs,
        )
        time_end = time.time()
        duration = time_end - time_start
        logger.debug(f"OD for orbit {orbit_id} completed in {duration:.3f}s.")
        od_orbits = qv.concatenate([od_orbits, od_orbit])
        if od_orbits.fragmented():
            od_orbits = qv.defragment(od_orbits)

        od_orbit_members = qv.concatenate([od_orbit_members, od_orbit_orbit_members])
        if od_orbit_members.fragmented():
            od_orbit_members = qv.defragment(od_orbit_members)

    return od_orbits, od_orbit_members


@ray.remote
def od_worker_remote(
    orbit_ids: npt.NDArray[np.str_],
    orbit_ids_indices: Tuple[int, int],
    orbits: FittedOrbits,
    orbit_members: FittedOrbitMembers,
    observations: OrbitDeterminationObservations,
    propagator: Type[Propagator],
    rchi2_threshold: float = 100,
    min_obs: int = 5,
    min_arc_length: float = 1.0,
    contamination_percentage: float = 0.0,
    delta: float = 1e-6,
    max_iter: int = 20,
    method: Literal["central", "finite"] = "central",
    propagator_kwargs: dict = {},
) -> Tuple[FittedOrbits, FittedOrbitMembers]:
    orbit_ids_chunk = orbit_ids[orbit_ids_indices[0] : orbit_ids_indices[1]]
    return od_worker(
        orbit_ids_chunk,
        orbits,
        orbit_members,
        observations,
        rchi2_threshold=rchi2_threshold,
        min_obs=min_obs,
        min_arc_length=min_arc_length,
        contamination_percentage=contamination_percentage,
        delta=delta,
        max_iter=max_iter,
        method=method,
        propagator=propagator,
        propagator_kwargs=propagator_kwargs,
    )


od_worker_remote.options(num_returns=1, num_cpus=1)


def od(
    orbit: FittedOrbits,
    observations: OrbitDeterminationObservations,
    propagator: Type[Propagator],
    rchi2_threshold: float = 100,
    min_obs: int = 5,
    min_arc_length: float = 1.0,
    contamination_percentage: float = 0.0,
    delta: float = 1e-6,
    max_iter: int = 20,
    method: Literal["central", "finite"] = "central",
    propagator_kwargs: dict = {},
) -> Tuple[FittedOrbits, FittedOrbitMembers]:
    # Intialize the propagator
    prop = propagator(**propagator_kwargs)

    if method not in ["central", "finite"]:
        err = "method should be one of 'central' or 'finite'."
        raise ValueError(err)

    obs_ids_all = observations.id.to_numpy(zero_copy_only=False)
    coords = observations.coordinates
    coords_sigma = coords.covariance.sigmas[:, 1:3]
    observers = observations.observers
    times_all = coords.time.mjd().to_numpy(zero_copy_only=False)

    # FLAG: can we stop iterating to find a solution?
    converged = False
    # FLAG: has an orbit with reduced chi2 less than the reduced chi2 of the input orbit been found?
    improved = False
    # FLAG: has an orbit with reduced chi2 less than the rchi2_threshold been found?
    solution_found = False
    # FLAG: is this orbit processable? Does it have at least min_obs observations?
    processable = True
    # FLAG: is this the first iteration with a successful differential correction (this solution is always stored as the solution
    # which needs to be improved.. input orbits may not have been previously corrected with current set of observations so this
    # forces at least one succesful iteration to have been taken.)
    first_solution = True

    num_obs = len(observations)
    if num_obs < min_obs:
        logger.debug("This orbit has fewer than {} observations.".format(min_obs))
        processable = False
    else:
        max_outliers = calculate_max_outliers(
            num_obs, min_obs, contamination_percentage
        )
        logger.debug(f"Maximum number of outliers allowed: {max_outliers}")
        outliers_tried = 0

        # Calculate chi2 for residuals on the given observations
        # for the current orbit, the goal is for the orbit to improve
        # such that the chi2 improves
        orbit_prev_ = orbit

        ephemeris_prev_ = prop.generate_ephemeris(
            orbit_prev_, observers, chunk_size=1, max_processes=1
        )

        # Calculate residuals and chi2
        residuals_prev_ = Residuals.calculate(
            coords,
            ephemeris_prev_.coordinates,
        )
        residuals_prev_array = np.stack(
            residuals_prev_.values.to_numpy(zero_copy_only=False)
        )[:, 1:3]

        num_obs_ = len(observations)
        chi2_prev_ = residuals_prev_.chi2.to_numpy()
        chi2_total_prev_ = np.sum(chi2_prev_)
        rchi2_prev_ = np.sum(chi2_prev_) / (2 * num_obs - 6)

        # Save the initial orbit in case we need to reset
        # to it later
        orbit_prev = orbit_prev_
        ephemeris_prev = ephemeris_prev_
        residuals_prev = residuals_prev_
        num_obs = num_obs_
        chi2_prev = chi2_prev_
        chi2_total_prev = chi2_total_prev_
        rchi2_prev = rchi2_prev_

        ids_mask = np.array([True for i in range(num_obs)])
        times_all = ephemeris_prev.coordinates.time.mjd().to_numpy()
        obs_id_outlier = []
        delta_prev = delta
        iterations = 0

        DELTA_INCREASE_FACTOR = 5
        DELTA_DECREASE_FACTOR = 100

        max_iter_i = max_iter
        max_iter_outliers = max_iter * (max_outliers + 1)

    while not converged and processable:
        iterations += 1

        # We add 1 here because the iterations are counted as they start, not
        # as they finish. There are a lot of 'continue' statements down below that
        # will exit the current iteration if something fails which makes accounting for
        # iterations at the start of an iteration easier.
        if iterations == max_iter_outliers + 1:
            logger.debug("Maximum number of iterations completed.")
            break
        if iterations == max_iter_i + 1 and (
            solution_found or (max_outliers == outliers_tried)
        ):
            logger.debug("Maximum number of iterations completed.")
            break
        logger.debug(f"Starting iteration number: {iterations}/{max_iter_outliers}")

        # Make sure delta is well bounded
        if delta_prev < 1e-14:
            delta_prev *= DELTA_INCREASE_FACTOR
            logger.debug("Delta is too small, increasing.")
        elif delta_prev > 1e-2:
            delta_prev /= DELTA_DECREASE_FACTOR
            logger.debug("Delta is too large, decreasing.")
        else:
            pass

        delta_iter = delta_prev
        logger.debug(f"Starting iteration {iterations} with delta {delta_iter}.")

        # Initialize the partials derivatives matrix
        num_params = 6
        A = np.zeros((2, num_params, num_obs))
        ATWA = np.zeros((num_params, num_params, num_obs))
        ATWb = np.zeros((num_params, 1, num_obs))

        # Generate ephemeris with current nominal orbit
        ephemeris_nom = prop.generate_ephemeris(
            orbit_prev, observers, chunk_size=1, max_processes=1
        )

        # Modify each component of the state by a small delta
        d = np.zeros((1, 6))
        for i in range(num_params):
            # zero the delta vector
            d *= 0.0

            # x, y, z [au]: 0, 1, 2
            # vx, vy, vz [au per day]: 3, 4, 5
            if i < 3:
                delta_iter = delta_prev
                d[0, i] = orbit_prev.coordinates.values[0, i] * delta_iter
            elif i < 6:
                delta_iter = delta_prev
                d[0, i] = orbit_prev.coordinates.values[0, i] * delta_iter

            # Modify component i of the orbit by a small delta
            cartesian_elements_p = orbit_prev.coordinates.values + d[0, :6]
            orbit_iter_p = Orbits.from_kwargs(
                coordinates=CartesianCoordinates.from_kwargs(
                    x=cartesian_elements_p[:, 0],
                    y=cartesian_elements_p[:, 1],
                    z=cartesian_elements_p[:, 2],
                    vx=cartesian_elements_p[:, 3],
                    vy=cartesian_elements_p[:, 4],
                    vz=cartesian_elements_p[:, 5],
                    time=orbit_prev.coordinates.time,
                    origin=orbit_prev.coordinates.origin,
                    frame=orbit_prev.coordinates.frame,
                )
            )

            # Calculate the modified ephemerides
            ephemeris_mod_p = prop.generate_ephemeris(
                orbit_iter_p, observers, chunk_size=1, max_processes=1
            )

            delta_denom = d[0, i]
            if method == "central":
                # Modify component i of the orbit by a small delta
                cartesian_elements_n = orbit_prev.coordinates.values - d[0, :6]
                orbit_iter_n = Orbits.from_kwargs(
                    coordinates=CartesianCoordinates.from_kwargs(
                        x=cartesian_elements_n[:, 0],
                        y=cartesian_elements_n[:, 1],
                        z=cartesian_elements_n[:, 2],
                        vx=cartesian_elements_n[:, 3],
                        vy=cartesian_elements_n[:, 4],
                        vz=cartesian_elements_n[:, 5],
                        time=orbit_prev.coordinates.time,
                        origin=orbit_prev.coordinates.origin,
                        frame=orbit_prev.coordinates.frame,
                    )
                )

                # Calculate the modified ephemerides
                ephemeris_mod_n = prop.generate_ephemeris(
                    orbit_iter_n, observers, chunk_size=1, max_processes=1
                )

                delta_denom *= 2

            else:
                ephemeris_mod_n = ephemeris_nom

            residuals_mod = Residuals.calculate(
                ephemeris_mod_p.coordinates,
                ephemeris_mod_n.coordinates,
            )
            residuals_mod = np.stack(
                residuals_mod.values.to_numpy(zero_copy_only=False)
            )
            residuals_mod_array = residuals_mod[:, 1:3]

            for n in range(num_obs):
                try:
                    A[:, i : i + 1, n] = (
                        residuals_mod_array[ids_mask][n : n + 1].T / delta_denom
                    )
                except RuntimeError:
                    print(orbit_prev.orbit_id)

        for n in range(num_obs):
            W = np.diag(1 / coords_sigma[n] ** 2)
            ATWA[:, :, n] = A[:, :, n].T @ W @ A[:, :, n]
            ATWb[:, :, n] = A[:, :, n].T @ W @ residuals_prev_array[n : n + 1].T

        ATWA = np.sum(ATWA, axis=2)
        ATWb = np.sum(ATWb, axis=2)

        try:
            # Calculate the condition number of both matrices
            ATWA_condition = np.linalg.cond(ATWA)
            ATWb_condition = np.linalg.cond(ATWb)
        except np.linalg.LinAlgError:
            ATWA_condition = np.nan
            ATWb_condition = np.nan
            logger.debug(
                f"Matrix condition calculation failed for {orbit.orbit_id[0].as_py()}"
            )

        if (ATWA_condition > 1e15) or (ATWb_condition > 1e15):
            delta_prev /= DELTA_DECREASE_FACTOR
            continue
        if np.any(np.isnan(ATWA)) or np.any(np.isnan(ATWb)):
            delta_prev *= DELTA_INCREASE_FACTOR
            continue
        else:
            try:
                delta_state = solve(
                    ATWA,
                    ATWb,
                ).T
                covariance_matrix = np.linalg.inv(ATWA)
                variances = np.diag(covariance_matrix)
                if np.any(variances <= 0) or np.any(np.isnan(variances)):
                    delta_prev /= DELTA_DECREASE_FACTOR
                    logger.debug(
                        "Variances are negative, 0.0, or NaN. Discarding solution."
                    )
                    continue

                r_variances = variances[0:3]
                r_sigma = np.sqrt(np.sum(r_variances))
                r = orbit_prev.coordinates.r_mag
                if (r_sigma / r) > 1:
                    delta_prev /= DELTA_DECREASE_FACTOR
                    logger.debug(
                        "Covariance matrix is largely unconstrained. Discarding solution."
                    )
                    continue

                if np.any(np.isnan(covariance_matrix)):
                    delta_prev *= DELTA_INCREASE_FACTOR
                    logger.debug(
                        "Covariance matrix contains NaNs. Discarding solution."
                    )
                    continue

            except np.linalg.LinAlgError:
                delta_prev *= DELTA_INCREASE_FACTOR
                continue

        if np.linalg.norm(delta_state[:3]) < 1e-16:
            logger.debug("Change in state is less than 1e-16 au, discarding solution.")
            delta_prev *= DELTA_DECREASE_FACTOR
            continue
        if np.linalg.norm(delta_state[:3]) > 100:
            delta_prev /= DELTA_DECREASE_FACTOR
            logger.debug("Change in state is more than 100 au, discarding solution.")
            continue

        cartesian_elements = orbit_prev.coordinates.values + delta_state
        orbit_iter = Orbits.from_kwargs(
            orbit_id=orbit_prev.orbit_id,
            coordinates=CartesianCoordinates.from_kwargs(
                x=cartesian_elements[:, 0],
                y=cartesian_elements[:, 1],
                z=cartesian_elements[:, 2],
                vx=cartesian_elements[:, 3],
                vy=cartesian_elements[:, 4],
                vz=cartesian_elements[:, 5],
                covariance=CoordinateCovariances.from_matrix(
                    covariance_matrix.reshape(1, 6, 6)
                ),
                time=orbit_prev.coordinates.time,
                origin=orbit_prev.coordinates.origin,
                frame=orbit_prev.coordinates.frame,
            ),
        )
        if np.linalg.norm(orbit_iter.coordinates.v_mag) > 1:
            delta_prev *= DELTA_INCREASE_FACTOR
            logger.debug("Orbit is moving extraordinarily fast, discarding solution.")
            continue

        # Generate ephemeris with current nominal orbit
        ephemeris_iter = prop.generate_ephemeris(
            orbit_iter, observers, chunk_size=1, max_processes=1
        )

        residuals = Residuals.calculate(coords, ephemeris_iter.coordinates)
        chi2_iter = residuals.chi2.to_numpy()
        chi2_total_iter = np.sum(chi2_iter[ids_mask])
        rchi2_iter = chi2_total_iter / (2 * num_obs - num_params)
        arc_length = times_all[ids_mask].max() - times_all[ids_mask].min()

        # If the new orbit has lower residuals than the previous,
        # accept the orbit and continue iterating until max iterations has been
        # reached. Once max iterations have been reached and the orbit still has not converged
        # to an acceptable solution, try removing an observation as an outlier and iterate again.
        if (
            (rchi2_iter < rchi2_prev) or first_solution
        ) and arc_length >= min_arc_length:
            if first_solution:
                logger.debug(
                    "Storing first successful differential correction iteration for these observations."
                )
                first_solution = False
            else:
                logger.debug("Potential improvement orbit has been found.")
            orbit_prev = orbit_iter
            residuals_prev = residuals
            chi2_prev = chi2_iter
            chi2_total_prev = chi2_total_iter
            rchi2_prev = rchi2_iter

            if rchi2_prev <= rchi2_prev_:
                improved = True

            if rchi2_prev <= rchi2_threshold:
                logger.debug("Potential solution orbit has been found.")
                solution_found = True
                converged = True

        elif (
            max_outliers > 0
            and outliers_tried <= max_outliers
            and iterations > max_iter_i
            and not solution_found
        ):
            logger.debug("Attempting to identify possible outliers.")
            # Previous fits have failed, lets reset the current best fit orbit back to its original
            # state and re-run fitting, this time removing outliers
            orbit_prev = orbit_prev_
            ephemeris_prev = ephemeris_prev_
            residuals_prev = residuals_prev_
            num_obs = num_obs_
            chi2_prev = chi2_prev_
            chi2_total_prev = chi2_total_prev_
            rchi2_prev = rchi2_prev_
            delta_prev = delta

            # Select i highest observations that contribute to
            # chi2 (and thereby the residuals)
            remove = chi2_prev.argsort()[-(outliers_tried + 1) :]

            # Grab the obs_ids for these outliers
            obs_id_outlier = obs_ids_all[remove]
            num_obs = len(observations) - len(obs_id_outlier)
            ids_mask = np.isin(obs_ids_all, obs_id_outlier, invert=True)
            arc_length = times_all[ids_mask].max() - times_all[ids_mask].min()

            logger.debug("Possible outlier(s): {}".format(obs_id_outlier))
            outliers_tried += 1
            if arc_length >= min_arc_length:
                max_iter_i = max_iter * (outliers_tried + 1)
            else:
                logger.debug(
                    "Removing the outlier will cause the arc length to go below the minimum."
                )

        # If the new orbit does not have lower residuals, try changing
        # delta to see if we get an improvement
        else:
            # logger.debug("Orbit did not improve since previous iteration, decrease delta and continue.")
            # delta_prev /= DELTA_DECREASE_FACTOR
            pass

        logger.debug(
            "Current r-chi2: {}, Previous r-chi2: {}, Max Iterations: {}, Outliers Tried: {}".format(
                rchi2_iter, rchi2_prev, max_iter_i, outliers_tried
            )
        )

    logger.debug("Solution found: {}".format(solution_found))
    logger.debug("First solution: {}".format(first_solution))

    if not solution_found or not processable or first_solution:
        od_orbit = FittedOrbits.empty()
        od_orbit_members = FittedOrbitMembers.empty()

    else:
        obs_times = observations.coordinates.time.mjd().to_numpy(zero_copy_only=False)[
            ids_mask
        ]
        arc_length_ = obs_times.max() - obs_times.min()
        assert arc_length == arc_length_

        od_orbit = FittedOrbits.from_kwargs(
            orbit_id=orbit_prev.orbit_id,
            object_id=orbit_prev.object_id,
            coordinates=orbit_prev.coordinates,
            arc_length=[arc_length_],
            num_obs=[num_obs],
            chi2=[chi2_total_prev],
            reduced_chi2=[rchi2_prev],
            iterations=[iterations],
            success=[improved],
            status_code=[0],
        )

        od_orbit_members = FittedOrbitMembers.from_kwargs(
            orbit_id=np.full(
                len(obs_ids_all), orbit_prev.orbit_id[0].as_py(), dtype="object"
            ),
            obs_id=obs_ids_all,
            residuals=residuals_prev,
            solution=np.isin(obs_ids_all, obs_id_outlier, invert=True),
            outlier=np.isin(obs_ids_all, obs_id_outlier),
        )

    return od_orbit, od_orbit_members


def differential_correction(
    orbits: Union[FittedOrbits, ray.ObjectRef],
    orbit_members: Union[FittedOrbitMembers, ray.ObjectRef],
    observations: Union[OrbitDeterminationObservations, ray.ObjectRef],
    propagator: Type[Propagator],
    min_obs: int = 5,
    min_arc_length: float = 1.0,
    contamination_percentage: float = 20,
    rchi2_threshold: float = 100,
    delta: float = 1e-8,
    max_iter: int = 20,
    method: Literal["central", "finite"] = "central",
    propagator_kwargs: dict = {},
    chunk_size: int = 10,
    max_processes: Optional[int] = 1,
    orbit_ids: Optional[npt.NDArray[np.str_]] = None,
    obs_ids: Optional[npt.NDArray[np.str_]] = None,
) -> Tuple[FittedOrbits, FittedOrbitMembers]:
    """
    Differentially correct (via finite/central differencing).

    Parameters
    ----------
    chunk_size : int, optional
        Number of orbits to send to each job.
    num_jobs : int, optional
        Number of jobs to launch.
    parallel_backend : str, optional
        Which parallelization backend to use {'ray', 'mp', 'cf'}. Defaults to using Python's concurrent.futures
        module ('cf').
    """
    time_start = time.perf_counter()
    logger.info("Running differential correction...")
    if isinstance(orbits, ray.ObjectRef):
        orbits_ref = orbits
        orbits = ray.get(orbits)
        logger.info("Retrieved orbits from the object store.")

        if orbit_ids is not None:
            orbits = orbits.apply_mask(pc.is_in(orbits.orbit_id, orbit_ids))
            logger.info("Applied mask to orbit members.")
    else:
        orbits_ref = None

    if isinstance(orbit_members, ray.ObjectRef):
        orbit_members_ref = orbit_members
        orbit_members = ray.get(orbit_members)
        logger.info("Retrieved orbit members from the object store.")

        if obs_ids is not None:
            orbit_members = orbit_members.apply_mask(
                pc.is_in(orbit_members.obs_id, obs_ids)
            )
            logger.info("Applied mask to orbit members.")
        if orbit_ids is not None:
            orbit_members = orbit_members.apply_mask(
                pc.is_in(orbit_members.orbit_id, orbit_ids)
            )
            logger.info("Applied mask to orbit members.")
    else:
        orbit_members_ref = None

    if isinstance(observations, ray.ObjectRef):
        observations_ref = observations
        observations = ray.get(observations)
        logger.info("Retrieved observations from the object store.")

        if obs_ids is not None:
            observations = observations.apply_mask(pc.is_in(observations.id, obs_ids))
            logger.info("Applied mask to observations.")
    else:
        observations_ref = None

    if len(orbits) == 0 or len(orbit_members) == 0:
        logger.info("Received no orbits or orbit members.")
        od_orbits = FittedOrbits.empty()
        od_orbit_members = FittedOrbitMembers.empty()
        time_end = time.perf_counter()
        logger.info(f"Differentially corrected {len(od_orbits)} orbits.")
        logger.info(
            f"Differential correction completed in {time_end - time_start:.3f} seconds."
        )
        return od_orbits, od_orbit_members

    orbit_ids = orbits.orbit_id.to_numpy(zero_copy_only=False)

    od_orbits = FittedOrbits.empty()
    od_orbit_members = FittedOrbitMembers.empty()

    if max_processes is None:
        max_processes = mp.cpu_count()

    use_ray = initialize_use_ray(num_cpus=max_processes)
    if use_ray:
        refs_to_free = []

        orbit_ids_ref = ray.put(orbit_ids)
        orbit_ids = ray.get(orbit_ids_ref)
        refs_to_free.append(orbit_ids_ref)
        logger.info("Placed orbit IDs in the object store.")

        if orbits_ref is None:
            orbits_ref = ray.put(orbits)
            orbits = ray.get(orbits_ref)
            refs_to_free.append(orbits_ref)
            logger.info("Placed orbits in the object store.")

        if orbit_members_ref is None:
            orbit_members_ref = ray.put(orbit_members)
            orbit_members = ray.get(orbit_members_ref)
            refs_to_free.append(orbit_members_ref)
            logger.info("Placed orbit members in the object store.")

        if observations_ref is None:
            observations_ref = ray.put(observations)
            refs_to_free.append(observations_ref)
            observations = ray.get(observations_ref)
            logger.info("Placed observations in the object store.")

        futures = []
        for orbit_ids_indices in _iterate_chunk_indices(orbit_ids, chunk_size):
            futures.append(
                od_worker_remote.remote(
                    orbit_ids_ref,
                    orbit_ids_indices,
                    orbits_ref,
                    orbit_members_ref,
                    observations_ref,
                    rchi2_threshold=rchi2_threshold,
                    min_obs=min_obs,
                    min_arc_length=min_arc_length,
                    contamination_percentage=contamination_percentage,
                    delta=delta,
                    max_iter=max_iter,
                    method=method,
                    propagator=propagator,
                    propagator_kwargs=propagator_kwargs,
                )
            )

            if len(futures) >= max_processes * 1.5:
                finished, futures = ray.wait(futures, num_returns=1)
                od_orbits_chunk, od_orbit_members_chunk = ray.get(finished[0])
                od_orbits = qv.concatenate([od_orbits, od_orbits_chunk])
                if od_orbits.fragmented():
                    od_orbits = qv.defragment(od_orbits)
                od_orbit_members = qv.concatenate(
                    [od_orbit_members, od_orbit_members_chunk]
                )
                if od_orbit_members.fragmented():
                    od_orbit_members = qv.defragment(od_orbit_members)

        while futures:
            finished, futures = ray.wait(futures, num_returns=1)
            od_orbits_chunk, od_orbit_members_chunk = ray.get(finished[0])
            od_orbits = qv.concatenate([od_orbits, od_orbits_chunk])
            if od_orbits.fragmented():
                od_orbits = qv.defragment(od_orbits)
            od_orbit_members = qv.concatenate(
                [od_orbit_members, od_orbit_members_chunk]
            )
            if od_orbit_members.fragmented():
                od_orbit_members = qv.defragment(od_orbit_members)

        if len(refs_to_free) > 0:
            ray.internal.free(refs_to_free)
            logger.info(
                f"Removed {len(refs_to_free)} references from the object store."
            )

    else:
        for orbit_ids_chunk in _iterate_chunks(orbit_ids, chunk_size):
            od_orbits_chunk, od_orbit_members_chunk = od_worker(
                orbit_ids_chunk,
                orbits,
                orbit_members,
                observations,
                rchi2_threshold=rchi2_threshold,
                min_obs=min_obs,
                min_arc_length=min_arc_length,
                contamination_percentage=contamination_percentage,
                delta=delta,
                max_iter=max_iter,
                method=method,
                propagator=propagator,
                propagator_kwargs=propagator_kwargs,
            )
            od_orbits = qv.concatenate([od_orbits, od_orbits_chunk])
            if od_orbits.fragmented():
                od_orbits = qv.defragment(od_orbits)
            od_orbit_members = qv.concatenate(
                [od_orbit_members, od_orbit_members_chunk]
            )
            if od_orbit_members.fragmented():
                od_orbit_members = qv.defragment(od_orbit_members)

    time_end = time.perf_counter()
    logger.info(f"Differentially corrected {len(od_orbits)} orbits.")
    logger.info(
        f"Differential correction completed in {time_end - time_start:.3f} seconds."
    )

    return od_orbits, od_orbit_members



from typing import Any, Dict, List, Literal, Union

import numpy as np
import numpy.typing as npt
import quivr as qv
import requests

from adam_core.coordinates.covariances import _upper_triangular_to_full

from ...coordinates import CoordinateCovariances, KeplerianCoordinates, Origin
from ...orbits import Orbits
from ...time import Timestamp


def _parse_oef(data: str) -> Dict[str, Any]:
    """
    Parse a OEF file and return the stored orbital elements.

    Parameters
    ----------
    data: str
        The content of the OEF file.

    Returns
    -------
    Dict[str, Any]
        Dictionary containing the parsed orbital elements and metadata.


    Examples
    --------
    format  = 'OEF2.0'       ! file format
    rectype = 'ML'           ! record type (1L/ML)
    refsys  = ECLM J2000     ! default reference system
    END_OF_HEADER
    2024YR4
    ! Keplerian elements: a, e, i, long. node, arg. peric., mean anomaly
     KEP   2.5158127507489616E+00  6.6154036821914619E-01   3.4081393687180 271.3655954496424 134.3614240204325  4.0403920526717883E+01
     MJD     60800.000000000 TDT
     MAG  23.876  0.150
    ! Non-gravitational parameters: model used, number of model parameters, dimension
     LSP   0  0    6
    ! PERIHELION   8.5150105724807035E-01
    ! APHELION   4.1801244442498522E+00
    ! ANODE   1.6132920678553648E+00
    ! DNODE  -1.6139338737144644E-02
    ! MOID   2.8281976977061222E-03
    ! PERIOD   1.4575246142278593E+03
    ! PHA F
    ! VINFTY    14.2161102117779
    ! U_PAR 5.5
    ! ORB_TYPE Apollo
    ! RMS    1.45945E-04   2.08511E-05   7.80533E-05   1.08159E-05   9.07220E-05   3.57225E-03
     COV   2.129990103626278E-08  3.043103695236090E-09  1.138994073085263E-08
     COV  -1.297300567885438E-09 -1.321094812357632E-08 -5.213516734886736E-07
     COV   4.347664336862408E-10  1.627274457493249E-09 -1.853526029216412E-10
     COV  -1.887473042074563E-09 -7.448518590163292E-08  6.092321667952164E-09
     COV  -6.879908552449990E-10 -7.069797588501348E-09 -2.787885242390572E-07
     COV   1.169847013231039E-10  7.781995171805923E-10  3.175091792990313E-08
     COV   8.230474520730192E-09  3.233601008844550E-07  1.276097850815426E-05
     COR   1.000000000000000E+00  9.999998967955239E-01  9.998647520507938E-01
     COR  -8.218400094356861E-01 -9.977753189147101E-01 -9.999999781092901E-01
     COR   9.999999999999999E-01  9.998650578225570E-01 -8.218757198089662E-01
     COR  -9.977926872410703E-01 -9.999998019539900E-01  9.999999999999998E-01
     COR  -8.149420314930980E-01 -9.983966836512449E-01 -9.998652884032685E-01
     COR   1.000000000000000E+00  7.930744967922404E-01  8.217690139449345E-01
     COR   1.000000000000000E+00  9.977735927640640E-01  1.000000000000000E+00

    """
    lines = data.strip().split("\n")
    result = {}

    # Parse header
    header = {}
    for line in lines:
        line = line.strip()
        if line == "END_OF_HEADER":
            break
        if "=" in line:
            key, value = line.split("=", 1)
            header[key.strip()] = value.split("!")[0].strip().strip("'")
    result["header"] = header

    # Find object ID
    for line in lines:
        if not line.startswith(
            ("!", " ", "format", "rectype", "refsys", "END_OF_HEADER")
        ):
            result["object_id"] = line.strip()
            break

    # Parse Keplerian elements
    for line in lines:
        if line.strip().startswith("KEP"):
            elements = line.split()[1:]
            result["elements"] = {
                "a": float(elements[0]),  # semi-major axis
                "e": float(elements[1]),  # eccentricity
                "i": float(elements[2]),  # inclination
                "node": float(elements[3]),  # longitude of ascending node
                "peri": float(elements[4]),  # argument of perihelion
                "M": float(elements[5]),  # mean anomaly
            }

    # Parse epoch
    for line in lines:
        if line.strip().startswith("MJD"):
            result["epoch"] = float(line.split()[1])
            result["time_system"] = line.split()[2]

    # Parse magnitude
    for line in lines:
        if line.strip().startswith("MAG"):
            mag_data = line.split()[1:]
            result["magnitude"] = {
                "value": float(mag_data[0]),
                "uncertainty": float(mag_data[1]),
            }

    # Parse derived parameters (marked with !)
    derived = {}
    for line in lines:
        if line.strip().startswith("!") and len(line.split()) >= 3:
            key = line.split()[1].lower()
            try:
                value = float(line.split()[2])
                derived[key] = value
            except ValueError:
                derived[key] = line.split()[2]
    result["derived"] = derived

    # Parse covariance matrix
    cov_matrix = []
    for line in lines:
        if line.strip().startswith("COV"):
            cov_matrix.extend([float(x) for x in line.split()[1:]])
    if cov_matrix:
        # Upper triangular matrix with order
        # (1,1)   (1,2)   (1,3)
        # (1,4)   (1,5)   (1,6)
        # (2,2)   (2,3)   (2,4)
        # (2,5)   (2,6)   (3,3)
        # (3,4)   (3,5)   (3,6)
        # (4,4)   (4,5)   (4,6)
        # (5,5)   (5,6)   (6,6)
        result["covariance"] = _upper_triangular_to_full(np.array(cov_matrix))

    # Parse correlation matrix
    cor_matrix = []
    for line in lines:
        if line.strip().startswith("COR"):
            cor_matrix.extend([float(x) for x in line.split()[1:]])
    if cor_matrix:
        result["correlation"] = _upper_triangular_to_full(np.array(cor_matrix))

    return result


def query_neocc(
    object_ids: Union[List, npt.ArrayLike],
    orbit_type: Literal["ke", "eq"] = "ke",
    orbit_epoch: Literal["middle", "present-day"] = "present-day",
) -> Orbits:
    """
    Query ESA's Near-Earth Object Coordination Centre (NEOCC) database for orbital elements of the specified NEOs.

    Parameters
    ----------
    object_ids : Union[List, npt.ArrayLike]
        Object IDs / designations recognizable by NEOCC.
    orbit_type : ["ke", "eq"]
        Type of orbital elements to query.
    orbit_epoch : ["middle", "present-day"]
        Epoch of the orbital elements to query.

    Returns
    -------
    orbits : `~adam_core.orbits.Orbits`
        Orbits object containing the orbital elements of the specified NEOs.
    """
    base_url = "https://neo.ssa.esa.int/PSDB-portlet/download"

    if orbit_type == "eq":
        raise NotImplementedError("Equinoctial elements are not supported yet.")

    if orbit_epoch == "middle":
        orbit_epoch = 0
    elif orbit_epoch == "present-day":
        orbit_epoch = 1
    else:
        raise ValueError(f"Invalid orbit epoch: {orbit_epoch}")

    orbits = Orbits.empty()

    for object_id in object_ids:

        # Clean object ID so that there are no spaces
        object_id = object_id.replace(" ", "")

        params = {"file": f"{object_id}.{orbit_type}{orbit_epoch}"}

        response = requests.get(base_url, params=params)
        response.raise_for_status()

        data = _parse_oef(response.text)
        if orbit_type == "ke":

            time_scale = data["time_system"]
            if time_scale == "TDT":
                time_scale = "tt"
            else:
                raise ValueError(f"Unsupported time scale: {time_scale}")

            if data["header"]["refsys"] != "ECLM J2000":
                raise ValueError(
                    f"Unsupported reference system: {data['header']['refsys']}"
                )

            orbit = Orbits.from_kwargs(
                orbit_id=[data["object_id"]],
                object_id=[data["object_id"]],
                coordinates=KeplerianCoordinates.from_kwargs(
                    a=[data["elements"]["a"]],
                    e=[data["elements"]["e"]],
                    i=[data["elements"]["i"]],
                    raan=[data["elements"]["node"]],
                    ap=[data["elements"]["peri"]],
                    M=[data["elements"]["M"]],
                    time=Timestamp.from_mjd([data["epoch"]], scale=time_scale),
                    covariance=CoordinateCovariances.from_matrix(
                        data["covariance"].reshape(
                            1,
                            6,
                            6,
                        )
                    ),
                    frame="ecliptic",
                    origin=Origin.from_kwargs(code=["SUN"]),
                ).to_cartesian(),
            )
            orbits = qv.concatenate([orbits, orbit])

    return orbits



import numpy as np
import pyarrow as pa
import pytest

from ...coordinates.cartesian import CartesianCoordinates
from ...coordinates.origin import Origin
from ...observations.detections import PointSourceDetections
from ...observations.exposures import Exposures
from ...observers.observers import Observers
from ...orbits.physical_parameters import PhysicalParameters
from ...time import Timestamp
from ..absolute_magnitude import estimate_absolute_magnitude_v_from_detections
from ..bandpasses.api import map_to_canonical_filter_bands
from ..magnitude import predict_magnitudes


def _make_geometry(n: int) -> tuple[CartesianCoordinates, Observers]:
    time = Timestamp.from_mjd(np.full(n, 60000), scale="tdb")
    observer = Observers.from_kwargs(
        code=["500"] * n,
        coordinates=CartesianCoordinates.from_kwargs(
            x=np.full(n, 1.0),
            y=np.zeros(n),
            z=np.zeros(n),
            vx=np.zeros(n),
            vy=np.zeros(n),
            vz=np.zeros(n),
            time=time,
            frame="ecliptic",
            origin=Origin.from_kwargs(code=["SUN"] * n),
        ),
    )

    # Place objects at opposition-like geometry to keep expected behavior stable.
    obj = CartesianCoordinates.from_kwargs(
        x=np.full(n, 2.0),
        y=np.zeros(n),
        z=np.zeros(n),
        vx=np.zeros(n),
        vy=np.zeros(n),
        vz=np.zeros(n),
        time=time,
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN"] * n),
    )
    return obj, observer


def test_estimate_absolute_magnitude_missing_sigma(monkeypatch):
    # Two exposures, three detections (with duplication) to exercise exposure_id alignment.
    exp = Exposures.from_kwargs(
        id=["e1", "e2"],
        start_time=Timestamp.from_mjd([60000, 60000], scale="tdb"),
        duration=[0.0, 0.0],
        # Non-canonical reported bands (exercise find_suggested_filter_bands fallback).
        filter=["g", "r"],
        observatory_code=["X05", "X05"],
        seeing=[None, None],
        depth_5sigma=[None, None],
    )
    det = PointSourceDetections.from_kwargs(
        id=["d1", "d2", "d3"],
        exposure_id=["e1", "e2", "e1"],
        time=Timestamp.from_mjd([60000, 60000, 60000], scale="tdb"),
        ra=[0.0, 0.0, 0.0],
        dec=[0.0, 0.0, 0.0],
        mag=[None, None, None],
        mag_sigma=[None, None, None],
    )

    obj, observer = _make_geometry(n=len(det))

    def fake_observers(self, *args, **kwargs):  # noqa: ARG001
        return observer

    monkeypatch.setattr(Exposures, "observers", fake_observers)

    # Generate synthetic observed magnitudes using the same forward model.
    exp_aligned = exp.take(pa.array([0, 1, 0], type=pa.int32()))
    canonical = map_to_canonical_filter_bands(
        exp_aligned.observatory_code, exp_aligned.filter
    )
    exp_canon = exp_aligned.set_column(
        "filter", pa.array(canonical.tolist(), type=pa.large_string())
    )

    H_true = 18.7
    mags_true = predict_magnitudes(
        H_true, obj, exp_canon, reference_filter="V", composition="NEO"
    )

    rng = np.random.default_rng(123)
    noise_sigma = 0.12
    mags_obs = np.asarray(mags_true, dtype=np.float64) + rng.normal(
        0.0, noise_sigma, size=len(det)
    )
    det = det.set_column("mag", mags_obs)

    est = estimate_absolute_magnitude_v_from_detections(
        det, exp, obj, composition="NEO", G=0.15, strict_band_mapping=False
    )

    assert isinstance(est, PhysicalParameters)
    assert len(est) == 1
    assert float(est.H_v[0].as_py()) == pytest.approx(H_true, abs=5e-2)
    assert est.H_v_sigma[0].as_py() is not None
    assert est.sigma_eff[0].as_py() is not None
    assert est.chi2_red[0].as_py() is None


def test_estimate_absolute_magnitude_with_sigma(monkeypatch):
    exp = Exposures.from_kwargs(
        id=["e1", "e2"],
        start_time=Timestamp.from_mjd([60000, 60000], scale="tdb"),
        duration=[0.0, 0.0],
        # Canonical filter IDs (no mapping needed, but mapping should be pass-through).
        filter=["LSST_r", "DECam_g"],
        observatory_code=["X05", "W84"],
        seeing=[None, None],
        depth_5sigma=[None, None],
    )
    det = PointSourceDetections.from_kwargs(
        id=["d1", "d2", "d3", "d4"],
        exposure_id=["e1", "e2", "e2", "e1"],
        time=Timestamp.from_mjd([60000, 60000, 60000, 60000], scale="tdb"),
        ra=[0.0, 0.0, 0.0, 0.0],
        dec=[0.0, 0.0, 0.0, 0.0],
        mag=[None, None, None, None],
        mag_sigma=[None, None, None, None],
    )

    obj, observer = _make_geometry(n=len(det))

    def fake_observers(self, *args, **kwargs):  # noqa: ARG001
        return observer

    monkeypatch.setattr(Exposures, "observers", fake_observers)

    exp_aligned = exp.take(pa.array([0, 1, 1, 0], type=pa.int32()))
    H_true = 21.3
    mags_true = predict_magnitudes(
        H_true, obj, exp_aligned, reference_filter="V", composition="C"
    )

    sigma = np.full(len(det), 0.1, dtype=np.float64)
    rng = np.random.default_rng(456)
    mags_obs = np.asarray(mags_true, dtype=np.float64) + rng.normal(
        0.0, sigma, size=len(det)
    )
    det = det.set_column("mag", mags_obs).set_column("mag_sigma", sigma)

    est = estimate_absolute_magnitude_v_from_detections(
        det, exp, obj, composition="C", G=0.15, strict_band_mapping=False
    )

    assert isinstance(est, PhysicalParameters)
    assert len(est) == 1
    assert float(est.H_v[0].as_py()) == pytest.approx(H_true, abs=7e-2)
    assert est.chi2_red[0].as_py() is not None
    assert est.H_v_sigma[0].as_py() is not None



from dataclasses import dataclass
from typing import Optional, Union

from .lua import LuaDict


@dataclass(kw_only=True)
class Translation(LuaDict):
    type: str


@dataclass(kw_only=True)
class Transform(LuaDict):
    translation: Translation


@dataclass(kw_only=True)
class KeplerTranslation(Translation):
    ### See: https://docs.openspaceproject.com/latest/reference/asset-components/KeplerTranslation.html
    argument_of_periapsis: float
    ascending_node: float
    eccentricity: float
    epoch: str
    inclination: float
    mean_anomaly: float
    period: float
    semi_major_axis: float
    type: str = "KeplerTranslation"


@dataclass(kw_only=True)
class SpiceTranslation(Translation):
    ### See: https://docs.openspaceproject.com/latest/reference/asset-components/SpiceTranslation.html
    observer: Union[str, int]
    target: Union[str, int]
    fixed_date: Optional[str] = None
    frame: Optional[str] = None
    type: str = "SpiceTranslation"



import logging

import numpy.typing as npt
import pyarrow as pa
import pyarrow.compute as pc
import quivr as qv
import requests

from ...coordinates.cometary import CometaryCoordinates
from ...coordinates.origin import Origin
from ...time import Timestamp
from ..variants import VariantOrbits

logger = logging.getLogger(__name__)


class ScoutObjectSummary(qv.Table):
    """
    Table the represents results of the Scout summary query (returns all objects)

    Those objects look like this
    {
      "unc": "59",
      "lastRun": "2024-10-18 15:31",
      "dec": "-19",
      "H": "18.7",
      "moid": "0.1",
      "geocentricScore": 0,
      "ra": "16:54",
      "rating": null,
      "tisserandScore": 15,
      "uncP1": "76",
      "ieoScore": 47,
      "rate": "2.2",
      "rmsN": "0.00",
      "Vmag": "20.5",
      "neoScore": 100,
      "nObs": 2,
      "objectName": "v17oct4",
      "phaScore": 20,
      "tEphem": "2024-10-22 20:15",
      "arc": "0.06",
      "caDist": null,
      "elong": "45",
      "vInf": null,
      "neo1kmScore": 5
    }
    """

    unc = qv.LargeStringColumn()
    lastRun = qv.LargeStringColumn()
    dec = qv.LargeStringColumn()
    H = qv.LargeStringColumn()
    moid = qv.LargeStringColumn()
    geocentricScore = qv.Int64Column()
    ra = qv.LargeStringColumn()
    rating = qv.Int8Column(nullable=True)
    tisserandScore = qv.Int64Column()
    uncP1 = qv.LargeStringColumn()
    ieoScore = qv.Int64Column()
    rate = qv.LargeStringColumn()
    rmsN = qv.LargeStringColumn()
    Vmag = qv.LargeStringColumn()
    neoScore = qv.Int64Column()
    nObs = qv.Int64Column()
    objectName = qv.LargeStringColumn()
    phaScore = qv.Int64Column()
    tEphem = qv.LargeStringColumn()
    arc = qv.LargeStringColumn()
    caDist = qv.LargeStringColumn(nullable=True)
    elong = qv.LargeStringColumn()
    vInf = qv.LargeStringColumn(nullable=True)
    neo1kmScore = qv.Int64Column()


def get_scout_objects() -> ScoutObjectSummary:
    """
    Query the Scout API for a summary of all objects.

    Returns
    -------
    scout_objects : `~adam_core.orbits.query.scout.ScoutObjectSummary`
        Table containing the summary of all objects.
    """
    url = "https://ssd-api.jpl.nasa.gov/scout.api"
    response = requests.get(url)
    response.raise_for_status()
    data = response.json()
    data = data["data"]
    table = pa.Table.from_pylist(data, schema=ScoutObjectSummary.schema)
    return ScoutObjectSummary.from_pyarrow(table)


class ScoutOrbit(qv.Table):
    """
    Represents a sample orbit from scout

    The fields are as follows:
    "fields":["idx","epoch","ec","qr","tp","om","w","inc","H","dca","tca","moid","vinf","geoEcc","impFlag"]
    Sample data:
          [
        0,
        "2457581.871499164",
        "3.357123709445450E-01",
        "9.083681207232809E-01",
        "2457636.871738402",
        "1.1141193497296813E+02",
        "2.4446138666195648E+02",
        "1.6610875455065550E+01",
        "24.694617",
        null,
        null,
        "0.321364995E-01",
        null,
        1.000000000E+99,
        0
      ],
    """

    idx = qv.Int64Column()
    epoch = qv.LargeStringColumn()
    ec = qv.LargeStringColumn()
    qr = qv.LargeStringColumn()
    tp = qv.LargeStringColumn()
    om = qv.LargeStringColumn()
    w = qv.LargeStringColumn()
    inc = qv.LargeStringColumn()
    H = qv.LargeStringColumn()
    dca = qv.LargeStringColumn(nullable=True)
    tca = qv.LargeStringColumn(nullable=True)
    moid = qv.LargeStringColumn()
    vinf = qv.LargeStringColumn(nullable=True)
    geoEcc = qv.LargeStringColumn()
    impFlag = qv.Int64Column()


def scout_orbits_to_variant_orbits(
    object_id: str, scout_orbits: ScoutOrbit
) -> VariantOrbits:
    """
    Convert a table of scout orbits to a table of variant orbits

    Parameters
    ----------
    scout_orbits : `~adam_core.orbits.query.scout.ScoutOrbit`
        Table containing the scout orbits

    Returns
    -------
    variant_orbits : `~adam_core.orbits.VariantOrbits`
        Table containing the variant orbits
    """
    cometary_coords = CometaryCoordinates.from_kwargs(
        q=pc.cast(scout_orbits.qr, pa.float64()),
        e=pc.cast(scout_orbits.ec, pa.float64()),
        i=pc.cast(scout_orbits.inc, pa.float64()),
        raan=pc.cast(scout_orbits.om, pa.float64()),
        ap=pc.cast(scout_orbits.w, pa.float64()),
        tp=pc.subtract(pc.cast(scout_orbits.tp, pa.float64()), 2400000.5),
        time=Timestamp.from_jd(pc.cast(scout_orbits.epoch, pa.float64())),
        origin=Origin.from_kwargs(code=pa.repeat("SUN", len(scout_orbits))),
        frame="ecliptic",
    )

    cartesian_coords = cometary_coords.to_cartesian()

    unique_orbit_ids = pc.cast(scout_orbits.idx, pa.large_string())

    variants = VariantOrbits.from_kwargs(
        coordinates=cartesian_coords,
        orbit_id=unique_orbit_ids,
        variant_id=unique_orbit_ids,
        object_id=pa.repeat(object_id, len(scout_orbits)),
    )

    return variants


def query_scout(ids: npt.ArrayLike) -> VariantOrbits:
    """
    Query the Scout API for a list of objects by id

    We return VariantOrbits for each orbit sample
    as the Scout API does not provide covariance information.

    These can be further reconstructed into an Orbit with
    covariance information using the `adam_core.orbits.Orbits` class.

    Parameters
    ----------
    ids : array-like
        List of object ids to query

    Returns
    -------
    orbits : `~adam_core.orbits.VariantOrbits`
        Table containing the orbits of the objects
    """
    url = "https://ssd-api.jpl.nasa.gov/scout.api?tdes={}&orbits=1"

    variant_orbits = VariantOrbits.empty()
    for object_id in ids:
        response = requests.get(url.format(object_id))
        response.raise_for_status()
        data = response.json()
        data = data["orbits"]["data"]

        # Convert from list of rows to list of columns
        data = list(map(list, zip(*data)))
        table = pa.Table.from_arrays(data, schema=ScoutOrbit.schema)
        scout_orbits = ScoutOrbit.from_pyarrow(table)
        object_variants = scout_orbits_to_variant_orbits(object_id, scout_orbits)
        variant_orbits = qv.concatenate([variant_orbits, object_variants])

    return variant_orbits



# flake8: noqa: F401
from .absolute_magnitude import estimate_absolute_magnitude_v_from_detections
from .magnitude import (
    calculate_apparent_magnitude_v,
    calculate_apparent_magnitude_v_and_phase_angle,
    calculate_phase_angle,
    convert_magnitude,
    predict_magnitudes,
)

__all__ = [
    # Simple magnitude system
    "calculate_apparent_magnitude_v",
    "calculate_apparent_magnitude_v_and_phase_angle",
    "calculate_phase_angle",
    "convert_magnitude",
    "predict_magnitudes",
    # Inverse magnitude system
    "estimate_absolute_magnitude_v_from_detections",
]



"""
Bandpass response curves and asteroid spectral templates.

This subpackage vendors (as package data) a curated set of:
- instrument filter response curves (throughput vs wavelength), normalized and stored in Parquet
- asteroid reflectance templates (C, S, and mixes like NEO/MBA), also stored in Parquet
- precomputed template×filter integrals used to derive color terms efficiently

The initial implementation is data-only: it is designed so `magnitude.py` can
later resolve `(observatory_code, reported_band)` -> canonical `filter_id` and use
precomputed integrals for higher-fidelity conversions, without changing behavior yet.
"""

from .api import (
    assert_filter_ids_have_curves,
    bandpass_color_terms,
    bandpass_delta_mag,
    compute_mix_integrals,
    get_integrals,
    load_asteroid_templates,
    load_bandpass_curves,
    load_observatory_band_map,
    load_template_integrals,
    map_to_canonical_filter_bands,
    register_custom_template,
)
from .tables import (
    AsteroidTemplates,
    BandpassCurves,
    ObservatoryBandMap,
    TemplateBandpassIntegrals,
)

__all__ = [
    # Tables
    "BandpassCurves",
    "ObservatoryBandMap",
    "AsteroidTemplates",
    "TemplateBandpassIntegrals",
    # Loaders
    "load_bandpass_curves",
    "load_observatory_band_map",
    "load_asteroid_templates",
    "load_template_integrals",
    # Utilities
    "map_to_canonical_filter_bands",
    "assert_filter_ids_have_curves",
    "get_integrals",
    "compute_mix_integrals",
    "bandpass_delta_mag",
    "bandpass_color_terms",
    "register_custom_template",
]



from __future__ import annotations

import numpy as np
import numpy.typing as npt
import pyarrow as pa
import pyarrow.compute as pc

from ..coordinates.cartesian import CartesianCoordinates
from ..observations.detections import PointSourceDetections
from ..observations.exposures import Exposures
from ..orbits.physical_parameters import PhysicalParameters
from .bandpasses.api import map_to_canonical_filter_bands
from .magnitude import predict_magnitudes
from .magnitude_common import BandpassComposition


def _as_float64_nan(a: pa.Array | pa.ChunkedArray) -> npt.NDArray[np.float64]:
    # Arrow nulls map to NaN in the output.
    return np.asarray(a.to_numpy(zero_copy_only=False), dtype=np.float64)


def _mad_sigma(x: npt.NDArray[np.float64]) -> float:
    # Robust scale estimate: 1.4826 * MAD for Gaussian-consistent sigma.
    med = float(np.nanmedian(x))
    mad = float(np.nanmedian(np.abs(x - med)))
    return 1.4826 * mad


def estimate_absolute_magnitude_v_from_detections(
    detections: PointSourceDetections,
    exposures: Exposures,
    object_coords: CartesianCoordinates,
    *,
    composition: BandpassComposition,
    G: float = 0.15,
    strict_band_mapping: bool = False,
    reference_filter: str = "V",
) -> PhysicalParameters:
    """
    Estimate V-band absolute magnitude H from observed apparent magnitudes.

    Assumptions
    -----------
    - Orbit has already been fit; `object_coords` are the heliocentric object coordinates at the
      observation times (aligned 1:1 with `detections`).
    - We estimate H only; `G` and `composition` are treated as fixed inputs.

    Parameters
    ----------
    detections
        Point-source detections. Uses `mag` and (optionally) `mag_sigma`.
    exposures
        Exposures referenced by `detections.exposure_id`. Uses `observatory_code` and `filter`.
    object_coords
        Object coordinates aligned with detections (same length and ordering).
    composition
        Required. Bandpass template ID (e.g. 'NEO') or (weight_C, weight_S) mix.
    G
        Fixed H-G slope parameter.
    strict_band_mapping
        If True, disallow SDSS/PS1 fallback filters when mapping reported bands.
    reference_filter
        Must be 'V' for this function.
    """
    if reference_filter != "V":
        raise ValueError(
            "reference_filter must be 'V' for estimate_absolute_magnitude_v_from_detections"
        )

    n_det = len(detections)
    if n_det == 0:
        raise ValueError("detections must be non-empty")
    if len(object_coords) != n_det:
        raise ValueError(
            f"object_coords length ({len(object_coords)}) must match detections length ({n_det})"
        )

    # ---------------------------------------------------------------------
    # Align exposures to detections order (vectorized join by exposure_id).
    # ---------------------------------------------------------------------
    if pc.any(pc.is_null(detections.exposure_id)).as_py():
        raise ValueError("detections.exposure_id must be non-null to link to exposures")

    idx = pc.fill_null(pc.index_in(detections.exposure_id, value_set=exposures.id), -1)
    idx_np = np.asarray(idx.to_numpy(zero_copy_only=False), dtype=np.int32)
    if np.any(idx_np < 0):
        missing = np.unique(
            np.asarray(
                detections.exposure_id.to_numpy(zero_copy_only=False), dtype=object
            )[idx_np < 0].astype(str)
        ).tolist()
        raise ValueError(f"detections reference unknown exposure_id(s): {missing}")

    exposures_aligned = exposures.take(pa.array(idx_np, type=pa.int32()))

    # Resolve (observatory_code, band/filter) -> canonical vendored filter_id.
    canonical = map_to_canonical_filter_bands(
        exposures_aligned.observatory_code,
        exposures_aligned.filter,
        allow_fallback_filters=not strict_band_mapping,
    )
    exposures_canon = exposures_aligned.set_column(
        "filter", pa.array(canonical.tolist(), type=pa.large_string())
    )

    # ---------------------------------------------------------------------
    # Forward model once at H=0 to get per-row offset, then solve H.
    # ---------------------------------------------------------------------
    m0 = predict_magnitudes(
        H=0.0,
        object_coords=object_coords,
        exposures=exposures_canon,
        G=G,
        reference_filter="V",
        composition=composition,
    )

    mag = _as_float64_nan(detections.mag)
    valid = np.isfinite(mag) & np.isfinite(m0)
    n_used = int(np.count_nonzero(valid))
    if n_used == 0:
        raise ValueError(
            "no valid rows: need finite detections.mag and forward-model magnitudes"
        )

    H_i = mag[valid] - np.asarray(m0, dtype=np.float64)[valid]

    mag_sigma = _as_float64_nan(detections.mag_sigma)
    sigma_used = mag_sigma[valid]
    have_all_sigma = bool(np.all(np.isfinite(sigma_used)))

    if have_all_sigma:
        w = 1.0 / np.square(sigma_used)
        wsum = float(np.sum(w))
        if not np.isfinite(wsum) or wsum <= 0.0:
            raise ValueError("invalid weights derived from mag_sigma")
        H_hat = float(np.sum(w * H_i) / wsum)
    else:
        H_hat = float(np.mean(H_i))

    resid = H_i - H_hat

    sigma_eff = None
    if n_used >= 2:
        s = _mad_sigma(resid)
        if np.isfinite(s):
            sigma_eff = float(s)

    chi2_red = None
    H_sigma = None
    if have_all_sigma and n_used >= 2:
        w = 1.0 / np.square(sigma_used)
        chi2 = float(np.sum(w * np.square(resid)))
        chi2_red = chi2 / float(n_used - 1)
        H_sigma = float(np.sqrt(1.0 / np.sum(w)))
        if np.isfinite(chi2_red) and chi2_red > 1.0:
            H_sigma = float(H_sigma * np.sqrt(chi2_red))
    elif sigma_eff is not None and n_used >= 2:
        H_sigma = float(sigma_eff / np.sqrt(float(n_used)))

    # Represent this as a 1-row PhysicalParameters table so callers can directly attach
    # to `Orbits.physical_parameters`.
    return PhysicalParameters.from_kwargs(
        H_v=[H_hat],
        H_v_sigma=[H_sigma],
        G=[float(G)],
        G_sigma=[None],
        sigma_eff=[sigma_eff],
        chi2_red=[chi2_red],
    )



from pathlib import Path

import numpy as np

from ..neocc import _parse_oef, query_neocc

TESTDATA_DIR = Path(__file__).parent / "testdata" / "neocc"


def test__parse_oef_2024YR4_ke0():
    with open(TESTDATA_DIR / "2024YR4.ke0", "r") as file:
        data = file.read()

    result = _parse_oef(data)

    # Test header
    assert result["header"]["format"] == "OEF2.0"
    assert result["header"]["rectype"] == "ML"
    assert result["header"]["refsys"] == "ECLM J2000"

    # Test object identification
    assert result["object_id"] == "2024YR4"

    # Test Keplerian elements
    assert result["elements"]["a"] == 2.5164332860625849
    assert result["elements"]["e"] == 0.66161845913486128
    assert result["elements"]["i"] == 3.4083351351535
    assert result["elements"]["node"] == 271.3684549331765
    assert result["elements"]["peri"] == 134.3639170929742
    assert result["elements"]["M"] == 12.947786309204405

    # Test epoch and time system
    assert result["epoch"] == 60688.865642464
    assert result["time_system"] == "TDT"

    # Test magnitude
    assert result["magnitude"]["value"] == 23.876
    assert result["magnitude"]["uncertainty"] == 0.150

    # Test derived parameters
    assert abs(result["derived"]["perihelion"] - 0.85151457282218190) < 1e-14
    assert abs(result["derived"]["aphelion"] - 4.1813519993029882) < 1e-14
    assert abs(result["derived"]["period"] - 1458.0639039239943) < 1e-14
    assert result["derived"]["pha"] == "F"
    assert result["derived"]["orb_type"] == "Apollo"

    # Test covariance and correlation matrices
    assert result["covariance"].shape == (6, 6)
    assert result["correlation"].shape == (6, 6)

    # Test matrix properties
    assert np.allclose(result["covariance"], result["covariance"].T)  # Symmetry
    assert np.allclose(result["correlation"], result["correlation"].T)  # Symmetry
    assert np.allclose(
        np.diagonal(result["correlation"]), 1.0
    )  # Diagonal elements should be 1


def test__parse_oef_2022OB5_ke1():
    with open(TESTDATA_DIR / "2022OB5.ke1", "r") as file:
        data = file.read()

    result = _parse_oef(data)

    # Test header
    assert result["header"]["format"] == "OEF2.0"
    assert result["header"]["rectype"] == "ML"
    assert result["header"]["refsys"] == "ECLM J2000"

    # Test object identification
    assert result["object_id"] == "2022OB5"

    # Test Keplerian elements
    assert result["elements"]["a"] == 1.0088199810054714
    assert result["elements"]["e"] == 0.058593341916141808
    assert result["elements"]["i"] == 2.0599781391032
    assert result["elements"]["node"] == 302.6856493105631
    assert result["elements"]["peri"] == 105.7080782884594
    assert result["elements"]["M"] == 334.06138612130985

    # Test epoch and time system
    assert result["epoch"] == 60600.0
    assert result["time_system"] == "TDT"

    # Test magnitude
    assert result["magnitude"]["value"] == 28.813
    assert result["magnitude"]["uncertainty"] == 0.150

    # Test derived parameters
    assert abs(result["derived"]["perihelion"] - 0.94970984692658189) < 1e-14
    assert abs(result["derived"]["aphelion"] - 1.0679301150843603) < 1e-14
    assert abs(result["derived"]["period"] - 370.09987635676049) < 1e-14
    assert result["derived"]["pha"] == "F"
    assert result["derived"]["orb_type"] == "Apollo"

    # Test covariance and correlation matrices
    assert result["covariance"].shape == (6, 6)
    assert result["correlation"].shape == (6, 6)

    # Test matrix properties
    assert np.allclose(result["covariance"], result["covariance"].T)  # Symmetry
    assert np.allclose(result["correlation"], result["correlation"].T)  # Symmetry
    assert np.allclose(
        np.diagonal(result["correlation"]), 1.0
    )  # Diagonal elements should be 1


def test_query_neocc(mocker):
    """Test query_neocc with both present-day and middle epochs (using saved test data)"""
    import requests

    # Setup mock responses for each test file
    mock_responses = {}
    for orbit_type in ["ke0", "ke1"]:
        for obj in ["2024YR4", "2022OB5"]:
            with open(TESTDATA_DIR / f"{obj}.{orbit_type}", "r") as f:
                mock_responses[f"{obj}.{orbit_type}"] = f.read()

    # Create mock response
    def mock_get(url, params):
        mock = mocker.MagicMock()
        mock.status_code = 200
        mock.text = mock_responses[params["file"]]
        return mock

    # Patch requests.get
    mocker.patch("requests.get", side_effect=mock_get)

    # Test querying multiple objects
    object_ids = ["2024YR4", "2022OB5"]
    orbits = query_neocc(object_ids, orbit_type="ke", orbit_epoch="present-day")

    # Verify the results
    assert orbits is not None
    assert len(orbits) == 2

    assert orbits.orbit_id[0].as_py() == "2024YR4"
    assert orbits.object_id[0].as_py() == "2024YR4"
    assert orbits.orbit_id[1].as_py() == "2022OB5"
    assert orbits.object_id[1].as_py() == "2022OB5"
    assert orbits.coordinates.time.days[0].as_py() == 60800
    assert orbits.coordinates.time.days[1].as_py() == 60600
    assert orbits.coordinates.time.scale == "tt"
    assert np.all(~np.isnan(orbits.coordinates.values))
    assert np.all(~np.isnan(orbits.coordinates.covariance.to_matrix()))

    # Verify the mock was called with correct parameters
    requests.get.assert_has_calls(
        [
            mocker.call(
                "https://neo.ssa.esa.int/PSDB-portlet/download",
                params={"file": "2024YR4.ke1"},
            ),
            mocker.call(
                "https://neo.ssa.esa.int/PSDB-portlet/download",
                params={"file": "2022OB5.ke1"},
            ),
        ]
    )

    # Test querying multiple objects
    object_ids = ["2024YR4", "2022OB5"]
    orbits = query_neocc(object_ids, orbit_type="ke", orbit_epoch="middle")

    # Verify the results
    assert orbits is not None
    assert len(orbits) == 2

    assert orbits.orbit_id[0].as_py() == "2024YR4"
    assert orbits.object_id[0].as_py() == "2024YR4"
    assert orbits.orbit_id[1].as_py() == "2022OB5"
    assert orbits.object_id[1].as_py() == "2022OB5"
    assert orbits.coordinates.time.days[0].as_py() == 60688
    assert orbits.coordinates.time.days[1].as_py() == 59792
    assert orbits.coordinates.time.scale == "tt"
    assert np.all(~np.isnan(orbits.coordinates.values))
    assert np.all(~np.isnan(orbits.coordinates.covariance.to_matrix()))

    # Verify the mock was called with correct parameters
    requests.get.assert_has_calls(
        [
            mocker.call(
                "https://neo.ssa.esa.int/PSDB-portlet/download",
                params={"file": "2024YR4.ke0"},
            ),
            mocker.call(
                "https://neo.ssa.esa.int/PSDB-portlet/download",
                params={"file": "2022OB5.ke0"},
            ),
        ]
    )



import numpy as np
import pyarrow.compute as pc
import pytest

from ...coordinates import CartesianCoordinates, Origin, SphericalCoordinates
from ...orbits.physical_parameters import PhysicalParameters
from ...time import Timestamp
from ...utils.helpers.orbits import make_real_orbits
from ..variants import VariantEphemeris, VariantOrbits


def test_VariantOrbits():

    # Get a sample of real orbits
    orbits = make_real_orbits(10)

    # Create a variant orbits object (expands the covariance matrix)
    # around the mean state
    variant_orbits = VariantOrbits.create(orbits)

    # For these 10 orbits this will select sigma-points so lets
    # check that the number of sigma-points is correct
    assert len(variant_orbits) == len(orbits) * 13

    # Now lets collapse the sigma-points back and see if we can reconstruct
    # the input covairance matrix
    collapsed_orbits = variant_orbits.collapse(orbits)

    # Check that the covariance matrices are close
    np.testing.assert_allclose(
        collapsed_orbits.coordinates.covariance.to_matrix(),
        orbits.coordinates.covariance.to_matrix(),
        rtol=0,
        atol=1e-14,
    )

    # Check that the orbit ids are the same
    np.testing.assert_equal(
        collapsed_orbits.orbit_id.to_numpy(zero_copy_only=False),
        orbits.orbit_id.to_numpy(zero_copy_only=False),
    )


def test_VariantOrbits_collapse_by_object_id():
    """Test that VariantOrbits.collapse_by_object_id correctly collapses variants into mean orbits."""

    # Create variant orbits with multiple objects, each having multiple variants
    variant_orbits = VariantOrbits.from_kwargs(
        orbit_id=["obj1", "obj1", "obj1", "obj2", "obj2", "obj2"],
        object_id=["obj1", "obj1", "obj1", "obj2", "obj2", "obj2"],
        variant_id=["0", "1", "2", "0", "1", "2"],
        physical_parameters=PhysicalParameters.from_kwargs(
            H_v=[15.0, 15.0, 15.0, 17.5, 17.5, 17.5],
            G=[0.15, 0.15, 0.15, 0.25, 0.25, 0.25],
        ),
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0, 1.1, 0.9, 2.0, 2.1, 1.9],
            y=[1.0, 1.1, 0.9, 2.0, 2.1, 1.9],
            z=[1.0, 1.1, 0.9, 2.0, 2.1, 1.9],
            vx=[0.1, 0.11, 0.09, 0.2, 0.21, 0.19],
            vy=[0.1, 0.11, 0.09, 0.2, 0.21, 0.19],
            vz=[0.1, 0.11, 0.09, 0.2, 0.21, 0.19],
            time=Timestamp.from_mjd([60000] * 6),
            origin=Origin.from_kwargs(code=["SUN"] * 6),
            frame="ecliptic",
        ),
    )

    # Collapse the variants
    collapsed = variant_orbits.collapse_by_object_id()

    # Check basic properties
    assert len(collapsed) == 2  # Should have one orbit per object
    assert set(collapsed.object_id.to_pylist()) == {"obj1", "obj2"}

    # Check that means are computed correctly for each object
    obj1 = collapsed.select("object_id", "obj1")
    obj2 = collapsed.select("object_id", "obj2")

    # Check means for obj1
    np.testing.assert_allclose(
        obj1.coordinates.values[0],
        np.array([1.0, 1.0, 1.0, 0.1, 0.1, 0.1]),  # Expected mean for obj1
        rtol=1e-14,
    )

    # Check means for obj2
    np.testing.assert_allclose(
        obj2.coordinates.values[0],
        np.array([2.0, 2.0, 2.0, 0.2, 0.2, 0.2]),  # Expected mean for obj2
        rtol=1e-14,
    )

    # Physical parameters should be preserved per object_id
    assert obj1.physical_parameters.H_v[0].as_py() == 15.0
    assert obj1.physical_parameters.G[0].as_py() == 0.15
    assert obj2.physical_parameters.H_v[0].as_py() == 17.5
    assert obj2.physical_parameters.G[0].as_py() == 0.25

    # Check that covariance matrices are computed correctly
    # For obj1, the variance should be approximately 0.00667 for each component
    obj1_cov = obj1.coordinates.covariance.to_matrix()[0]
    # The variance is sum((x - mean)^2) / n where n=3
    # For positions: (1.1 - 1.0)^2 + (0.9 - 1.0)^2 + (1.0 - 1.0)^2 = 0.02
    # So variance = 0.02/3 ≈ 0.00667
    expected_variance_obj1_pos = 0.02 / 3  # Population variance
    expected_variance_obj1_vel = 0.0002 / 3  # Population variance
    np.testing.assert_allclose(
        np.diag(obj1_cov),
        [expected_variance_obj1_pos] * 3 + [expected_variance_obj1_vel] * 3,
        rtol=1e-6,
    )

    # For obj2, the variance should also be approximately 0.00667
    obj2_cov = obj2.coordinates.covariance.to_matrix()[0]
    expected_variance_obj2_pos = (
        0.02 / 3
    )  # (2.1 - 2.0)^2 + (1.9 - 2.0)^2 + (2.0 - 2.0)^2 = 0.02/3
    expected_variance_obj2_vel = (
        0.0002 / 3
    )  # (0.21 - 0.2)^2 + (0.19 - 0.2)^2 + (0.2 - 0.2)^2 = 0.0002/3
    np.testing.assert_allclose(
        np.diag(obj2_cov),
        [expected_variance_obj2_pos] * 3 + [expected_variance_obj2_vel] * 3,
        rtol=1e-6,
    )

    # Test that time and origin are preserved
    assert all(t == 60000 for t in collapsed.coordinates.time.mjd().to_pylist())
    assert all(o == "SUN" for o in collapsed.coordinates.origin.code.to_pylist())
    assert collapsed.coordinates.frame == "ecliptic"

    # Test error cases
    # Test that variants with different times raise an error
    variant_orbits_diff_times = VariantOrbits.from_kwargs(
        orbit_id=["obj1", "obj1"],
        object_id=["obj1", "obj1"],
        variant_id=["0", "1"],
        physical_parameters=PhysicalParameters.from_kwargs(
            H_v=[15.0, 15.0], G=[0.15, 0.15]
        ),
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0, 1.1],
            y=[1.0, 1.1],
            z=[1.0, 1.1],
            vx=[0.1, 0.11],
            vy=[0.1, 0.11],
            vz=[0.1, 0.11],
            time=Timestamp.from_mjd([60000, 60001]),  # Different times
            origin=Origin.from_kwargs(code=["SUN", "SUN"]),
            frame="ecliptic",
        ),
    )
    with pytest.raises(AssertionError):
        variant_orbits_diff_times.collapse_by_object_id()

    # Test that variants with different origins raise an error
    variant_orbits_diff_origins = VariantOrbits.from_kwargs(
        orbit_id=["obj1", "obj1"],
        object_id=["obj1", "obj1"],
        variant_id=["0", "1"],
        physical_parameters=PhysicalParameters.from_kwargs(
            H_v=[15.0, 15.0], G=[0.15, 0.15]
        ),
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0, 1.1],
            y=[1.0, 1.1],
            z=[1.0, 1.1],
            vx=[0.1, 0.11],
            vy=[0.1, 0.11],
            vz=[0.1, 0.11],
            time=Timestamp.from_mjd([60000, 60000]),
            origin=Origin.from_kwargs(code=["SUN", "EARTH"]),  # Different origins
            frame="ecliptic",
        ),
    )
    with pytest.raises(AssertionError):
        variant_orbits_diff_origins.collapse_by_object_id()


def test_VariantEphemeris_collapse_by_object_id_single_epoch():
    """Test that VariantEphemeris.collapse_by_object_id collapses by object_id for a single epoch."""
    variant_ephemeris = VariantEphemeris.from_kwargs(
        orbit_id=["obj1", "obj1", "obj1", "obj2", "obj2", "obj2"],
        object_id=["obj1", "obj1", "obj1", "obj2", "obj2", "obj2"],
        variant_id=["0", "1", "2", "0", "1", "2"],
        predicted_magnitude_v=[20.0, 21.0, 19.0, 18.0, 18.5, 17.5],
        coordinates=SphericalCoordinates.from_kwargs(
            rho=[1.0, 1.1, 0.9, 2.0, 2.1, 1.9],
            lon=[1.0, 1.1, 0.9, 2.0, 2.1, 1.9],
            lat=[1.0, 1.1, 0.9, 2.0, 2.1, 1.9],
            vrho=[0.1, 0.11, 0.09, 0.2, 0.21, 0.19],
            vlon=[0.1, 0.11, 0.09, 0.2, 0.21, 0.19],
            vlat=[0.1, 0.11, 0.09, 0.2, 0.21, 0.19],
            time=Timestamp.from_mjd([60000] * 6),
            origin=Origin.from_kwargs(code=["500"] * 6),
            frame="equatorial",
        ),
    )

    collapsed = variant_ephemeris.collapse_by_object_id()

    assert len(collapsed) == 2
    assert set(collapsed.object_id.to_pylist()) == {"obj1", "obj2"}

    obj1 = collapsed.select("object_id", "obj1")
    obj2 = collapsed.select("object_id", "obj2")

    np.testing.assert_allclose(
        obj1.coordinates.values[0],
        np.array([1.0, 1.0, 1.0, 0.1, 0.1, 0.1]),
        rtol=1e-14,
    )
    np.testing.assert_allclose(
        obj2.coordinates.values[0],
        np.array([2.0, 2.0, 2.0, 0.2, 0.2, 0.2]),
        rtol=1e-14,
    )

    # Mean magnitudes should be propagated.
    assert obj1.predicted_magnitude_v[0].as_py() == pytest.approx(20.0)
    assert obj2.predicted_magnitude_v[0].as_py() == pytest.approx(18.0)

    # Check covariance diagonals (population variance with n=3).
    obj1_cov = obj1.coordinates.covariance.to_matrix()[0]
    expected_variance_pos = 0.02 / 3
    expected_variance_vel = 0.0002 / 3
    np.testing.assert_allclose(
        np.diag(obj1_cov),
        [expected_variance_pos] * 3 + [expected_variance_vel] * 3,
        rtol=1e-6,
    )

    obj2_cov = obj2.coordinates.covariance.to_matrix()[0]
    np.testing.assert_allclose(
        np.diag(obj2_cov),
        [expected_variance_pos] * 3 + [expected_variance_vel] * 3,
        rtol=1e-6,
    )

    # Time, origin, and frame should be preserved.
    assert all(t == 60000 for t in collapsed.coordinates.time.mjd().to_pylist())
    assert all(o == "500" for o in collapsed.coordinates.origin.code.to_pylist())
    assert collapsed.coordinates.frame == "equatorial"


def test_VariantEphemeris_collapse_by_object_id_groups_by_time_and_origin():
    """Collapse should be performed per (object_id, time, origin_code) group."""
    variant_ephemeris = VariantEphemeris.from_kwargs(
        orbit_id=["obj1"] * 6,
        object_id=["obj1"] * 6,
        variant_id=["0", "1", "0", "1", "0", "1"],
        coordinates=SphericalCoordinates.from_kwargs(
            rho=[1.0, 1.2, 10.0, 10.2, 100.0, 100.2],
            lon=[1.0, 1.2, 10.0, 10.2, 100.0, 100.2],
            lat=[1.0, 1.2, 10.0, 10.2, 100.0, 100.2],
            vrho=[0.1, 0.12, 1.0, 1.02, 10.0, 10.02],
            vlon=[0.1, 0.12, 1.0, 1.02, 10.0, 10.02],
            vlat=[0.1, 0.12, 1.0, 1.02, 10.0, 10.02],
            time=Timestamp.from_mjd([60000, 60000, 60001, 60001, 60000, 60000]),
            origin=Origin.from_kwargs(code=["500", "500", "500", "500", "X05", "X05"]),
            frame="equatorial",
        ),
    )

    collapsed = variant_ephemeris.collapse_by_object_id()

    # Keys: (60000, "500"), (60001, "500"), (60000, "X05")
    assert len(collapsed) == 3
    assert set(collapsed.coordinates.origin.code.to_pylist()) == {"500", "X05"}
    assert set(collapsed.coordinates.time.mjd().to_pylist()) == {60000, 60001}

    # Validate one group: mjd=60000, origin=500 -> mean of [1.0, 1.2] => 1.1 etc.
    group = (
        collapsed.select("coordinates.origin.code", "500")
        .select("coordinates.time.days", 60000)
        .select("coordinates.time.nanos", 0)
    )
    assert len(group) == 1
    np.testing.assert_allclose(
        group.coordinates.values[0],
        np.array([1.1, 1.1, 1.1, 0.11, 0.11, 0.11]),
        rtol=0,
        atol=1e-12,
    )


def test_VariantEphemeris_collapse_by_object_id_wraps_longitude():
    """Longitude is circular in degrees; mean should respect wrap-around near 0/360."""
    variant_ephemeris = VariantEphemeris.from_kwargs(
        orbit_id=["obj1", "obj1", "obj1"],
        object_id=["obj1", "obj1", "obj1"],
        variant_id=["0", "1", "2"],
        coordinates=SphericalCoordinates.from_kwargs(
            rho=[1.0, 1.0, 1.0],
            lon=[359.0, 1.0, 0.0],
            lat=[0.0, 0.0, 0.0],
            vrho=[0.0, 0.0, 0.0],
            vlon=[0.0, 0.0, 0.0],
            vlat=[0.0, 0.0, 0.0],
            time=Timestamp.from_mjd([60000] * 3),
            origin=Origin.from_kwargs(code=["500"] * 3),
            frame="equatorial",
        ),
    )

    collapsed = variant_ephemeris.collapse_by_object_id()
    assert len(collapsed) == 1
    lon = collapsed.coordinates.values[0][1]
    assert lon == pytest.approx(0.0, abs=1e-12)
    assert 0.0 <= lon < 360.0


def test_VariantEphemeris_collapse_by_object_id_partial_aberrated_raises():
    """Aberrated coordinates are ignored/dropped and regenerated after collapse."""
    variant_ephemeris = VariantEphemeris.from_kwargs(
        orbit_id=["obj1", "obj1"],
        object_id=["obj1", "obj1"],
        variant_id=["0", "1"],
        coordinates=SphericalCoordinates.from_kwargs(
            rho=[1.0, 1.1],
            lon=[1.0, 1.1],
            lat=[1.0, 1.1],
            vrho=[0.1, 0.11],
            vlon=[0.1, 0.11],
            vlat=[0.1, 0.11],
            time=Timestamp.from_mjd([60000] * 2),
            origin=Origin.from_kwargs(code=["500"] * 2),
            frame="equatorial",
        ),
        aberrated_coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0, None],
            y=[1.0, None],
            z=[1.0, None],
            vx=[0.1, None],
            vy=[0.1, None],
            vz=[0.1, None],
            time=Timestamp.from_mjd([60000] * 2),
            origin=Origin.from_kwargs(code=["SUN"] * 2),
            frame="ecliptic",
        ),
    )

    collapsed = variant_ephemeris.collapse_by_object_id()
    assert len(collapsed) == 1
    assert not pc.all(pc.is_null(collapsed.aberrated_coordinates.x)).as_py()
    assert not pc.all(pc.is_null(collapsed.light_time)).as_py()


def test_VariantEphemeris_collapse_by_object_id_aberrated_times_can_vary():
    """Collapsed aberrated emission times should be consistent with light_time."""
    variant_ephemeris = VariantEphemeris.from_kwargs(
        orbit_id=["obj1", "obj1", "obj1"],
        object_id=["obj1", "obj1", "obj1"],
        variant_id=["0", "1", "2"],
        coordinates=SphericalCoordinates.from_kwargs(
            rho=[1.0, 1.1, 0.9],
            lon=[1.0, 1.1, 0.9],
            lat=[1.0, 1.1, 0.9],
            vrho=[0.1, 0.11, 0.09],
            vlon=[0.1, 0.11, 0.09],
            vlat=[0.1, 0.11, 0.09],
            time=Timestamp.from_mjd([60000.0] * 3, scale="utc"),
            origin=Origin.from_kwargs(code=["500"] * 3),
            frame="equatorial",
        ),
    )

    collapsed = variant_ephemeris.collapse_by_object_id()
    assert len(collapsed) == 1

    coords_tdb = collapsed.coordinates.time.rescale("tdb")
    aberr_tdb = collapsed.aberrated_coordinates.time.rescale("tdb")
    delta_days, delta_nanos = coords_tdb.difference(aberr_tdb)
    fractional_days = pc.divide(delta_nanos, 86400 * 1e9)
    delta = pc.add(delta_days, fractional_days).to_numpy(zero_copy_only=False)
    lt = collapsed.light_time.to_numpy(zero_copy_only=False)
    np.testing.assert_allclose(delta, lt, atol=1e-6)



from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Tuple, Union

from .lua import LuaDict
from .translation import Translation


class RenderBinMode(Enum):
    BACKGROUND = "Background"
    OPAQUE = "Opaque"
    PREDEFERREDTRANSPARENT = "PreDeferredTransparent"
    OVERLAY = "Overlay"
    POSTDEFERREDTRANSPARENT = "PostDeferredTransparent"
    STICKER = "Sticker"


class RenderableOrbitalKeplerFormat(Enum):
    TLE = "TLE"
    OMM = "OMM"
    SBDB = "SBDB"


class RenderableOrbitalKeplerRendering(Enum):
    TRAIL = "Trail"
    POINT = "Point"
    POINTS_TRAILS = "PointsTrails"


class RenderableTrailRendering(Enum):
    LINES = "Lines"
    POINTS = "Points"
    LINES_POINTS = "Lines+Points"


@dataclass(kw_only=True)
class Resource(LuaDict):
    """Custom class to point to local resource files"""

    path: str

    def to_string(self, indent: int = 0):
        return f'asset.resource("{self.path}")'


@dataclass(kw_only=True)
class Renderable(LuaDict):
    ### See:https://docs.openspaceproject.com/latest/reference/asset-components/Renderable.html
    type: str

    dim_in_atmosphere: Optional[bool] = None
    enabled: Optional[bool] = None
    opacity: Optional[float] = None
    render_bin_mode: Optional[RenderBinMode] = None
    tag: Optional[Union[str, List[str]]] = None


@dataclass(kw_only=True)
class RenderableOrbitalKepler(Renderable):
    ### See: https://docs.openspaceproject.com/latest/reference/asset-components/RenderableOrbitalKepler.html
    color: Tuple[float, float, float]
    format: RenderableOrbitalKeplerFormat
    path: Union[str, Resource]
    segment_quality: int
    type: str = "RenderableOrbitalKepler"

    contiguous_mode: Optional[bool] = None
    enable_max_size: Optional[bool] = None
    enable_outline: Optional[bool] = None
    max_size: Optional[float] = None
    outline_color: Optional[Tuple[float, float, float]] = None
    outline_width: Optional[float] = None
    point_size_exponent: Optional[float] = None
    rendering: Optional[RenderableOrbitalKeplerRendering] = None
    render_size: Optional[int] = None
    start_render_idx: Optional[int] = None
    trail_fade: Optional[float] = None
    trail_width: Optional[float] = None


@dataclass(kw_only=True)
class RenderableTrailOrbit(Renderable):
    ### See: https://docs.openspaceproject.com/latest/reference/asset-components/RenderableTrailOrbit.html
    color: Tuple[float, float, float]
    period: float  # In days
    resolution: int
    translation: Translation
    type: str = "RenderableTrailOrbit"

    enable_fade: Optional[bool] = None
    line_fade_amount: Optional[float] = None
    line_length: Optional[float] = None
    line_width: Optional[float] = None
    point_size: Optional[int] = None
    rendering: Optional[RenderableTrailRendering] = None


@dataclass(kw_only=True)
class RenderableTrailTrajectory(Renderable):
    ### See: https://docs.openspaceproject.com/latest/reference/asset-components/RenderableTrailTrajectory.html
    color: Tuple[float, float, float]
    end_time: str
    start_time: str
    translation: Translation
    type: str = "RenderableTrailTrajectory"

    accurate_trail_positions: Optional[int] = None
    enable_fade: Optional[bool] = None
    enable_sweep_chunking: Optional[int] = None
    line_fade_amount: Optional[float] = None
    line_length: Optional[float] = None
    line_width: Optional[float] = None
    point_size: Optional[int] = None
    rendering: Optional[RenderableTrailRendering] = None
    sample_interval: Optional[float] = None
    show_full_trail: Optional[bool] = None
    sweep_chunk_size: Optional[int] = None
    time_stamp_subsample_factor: Optional[int] = None



from typing import Iterable, Sequence, Tuple


def _iterate_chunks(iterable: Sequence, chunk_size: int) -> Iterable:
    """
    Generator that yields chunks of size chunk_size from sized iterable
    (Sequence).

    Parameters
    ----------
    iterable : Sequence
        Iterable to chunk.
    chunk_size : int
        Size of chunks.

    Yields
    ------
    chunk : iterable
        Chunk of size chunk_size from iterable.
    """
    N = len(iterable)
    for i in range(0, N, chunk_size):
        yield iterable[i : i + chunk_size]


def _iterate_chunk_indices(
    iterable: Sequence, chunk_size: int
) -> Iterable[Tuple[int, int]]:
    """
    Generator that yields indices for chunks of size chunk_size from sized
    iterable (Sequence).

    Parameters
    ----------
    iterable : Sequence
        Iterable to chunk.
    chunk_size : int
        Size of chunks.

    Yields
    ------
    chunk : iterable
        Chunk of size chunk_size from iterable.
    """
    N = len(iterable)
    for i in range(0, N, chunk_size):
        yield i, min(i + chunk_size, N)



import logging
from typing import Callable, Optional, Tuple

import numpy as np
import numpy.typing as npt
import pyarrow as pa
import quivr as qv
from scipy.linalg import sqrtm
from scipy.stats import multivariate_normal

from .jacobian import calc_jacobian

logger = logging.getLogger(__name__)

COVARIANCE_FILL_VALUE = np.nan


def sigmas_to_covariances(sigmas: np.ndarray) -> np.ndarray:
    """
    Convert an array of standard deviations to an array of covariance matrices.
    Non-diagonal elements are set to zero.

    Parameters
    ----------
    sigmas : `numpy.ndarray` (N, D)
        Standard deviations for N coordinates in D dimensions.

    Returns
    -------
    covariances : `numpy.ndarray` (N, D, D)
        Covariance matrices for N coordinates in D dimensions.
    """
    D = sigmas.shape[1]
    identity = np.identity(D, dtype=sigmas.dtype)
    covariances = np.einsum("kj,ji->kij", sigmas**2, identity, order="C")
    return covariances


class CoordinateCovariances(qv.Table):
    # TODO: Would be interesting if the dimensionality can be generalized
    #      to D dimensions, so (N, D, D) instead of (N, 6, 6). We would be
    #      able to use this class for the covariance matrices of different
    #      measurments like projections (D = 4) and photometry (D = 1).

    values = qv.LargeListColumn(pa.float64(), nullable=True)
    # When fixed, we should revert to:
    # values = Column(pa.fixed_shape_tensor(pa.float64(), (6, 6)))

    @property
    def sigmas(self):
        cov_diag = np.diagonal(self.to_matrix(), axis1=1, axis2=2)
        sigmas = np.sqrt(cov_diag)
        return sigmas

    def to_matrix(self) -> np.ndarray:
        """
        Return the covariance matrices as a 3D array of shape (N, 6, 6).

        Returns
        -------
        covariances : `numpy.ndarray` (N, 6, 6)
            Covariance matrices for N coordinates in 6 dimensions.
        """
        # return self.values.combine_chunks().to_numpy_ndarray()
        values = self.values.to_numpy(zero_copy_only=False)

        # Try and see if all covariance matrices are None, if so return
        # an array of NaNs.
        try:
            if np.all(values == None):  # noqa: E711
                return np.full((len(self), 6, 6), np.nan)

        except ValueError as e:
            err_str = (
                "The truth value of an array with more than one element is ambiguous."
                " Use a.any() or a.all()"
            )
            if str(e) != err_str:
                raise e

        # Try to stack the values into a 3D array. If this works, then
        # all covariance matrices are the same size and we can return
        # the stacked matrices.
        try:
            cov = np.stack(values).reshape(-1, 6, 6)

        # If not then some of the arrays might be None. Lets loop through
        # the values and fill in the arrays that are missing (None) with NaNs.
        except ValueError as e:
            # If we don't get the error we expect, then raise it.
            if str(e) != "all input arrays must have the same shape":
                raise e
            else:
                for i in range(len(values)):
                    if values[i] is None:
                        values[i] = np.full(36, np.nan)

            # Try stacking again
            cov = np.stack(values).reshape(-1, 6, 6)

        return cov

    @classmethod
    def from_matrix(cls, covariances: np.ndarray) -> "CoordinateCovariances":
        """
        Create a Covariances object from a 3D array of covariance matrices.

        Parameters
        ----------
        covariances : `numpy.ndarray` (N, 6, 6)
            Covariance matrices for N coordinates in 6 dimensions.

        Returns
        -------
        covariances : `Covariances`
            Covariance matrices for N coordinates in 6 dimensions.

        Raises
        ------
        ValueError : If the covariance matrices are not (N, 6, 6)
        """
        # cov = pa.FixedShapeTensorArray.from_numpy_ndarray(covariances)
        if covariances.shape[1:] != (6, 6):
            raise ValueError(
                f"Covariance matrices should have shape (N, 6, 6) but got {covariances.shape}"
            )
        cov = covariances.flatten()
        offsets = np.arange(0, (len(covariances) + 1) * 36, 36, dtype=np.int64)
        return cls.from_kwargs(values=pa.LargeListArray.from_arrays(offsets, cov))

    @classmethod
    def from_sigmas(cls, sigmas: np.ndarray) -> "CoordinateCovariances":
        """
        Create a Covariances object from a 2D array of sigmas.

        Parameters
        ----------
        sigmas : `numpy.ndarray` (N, 6)
            Array of 1-sigma uncertainties for N coordinates in 6
            dimensions.

        Returns
        -------
        covariances : `Covariances`
            Covariance matrices with the diagonal elements set to the
            squares of the input sigmas.
        """
        return cls.from_matrix(sigmas_to_covariances(sigmas))

    @classmethod
    def nulls(cls, length: int) -> "CoordinateCovariances":
        """
        Create a Covariances object with all covariance matrix elements set to NaN.
        Parameters
        ----------
        length : `int`
            Number of coordinates.

        Returns
        -------
        covariances : `CoordinateCovariances`
            Covariance matrices for N coordinates in 6 dimensions.
        """
        return cls.from_kwargs(
            values=pa.ListArray.from_arrays(
                pa.array(np.arange(0, 36 * (length + 1), 36)),
                pa.nulls(36 * length, pa.float64()),
            )
        )

    def is_all_nan(self) -> bool:
        """
        Check if all covariance matrix values are NaN.

        Returns
        -------
        is_all_nan : bool
            True if all covariance matrix elements are NaN, False otherwise.
        """
        return np.all(np.isnan(self.to_matrix()))


def make_positive_semidefinite(
    cov: np.ndarray, semidef_tol: float = 1e-15
) -> np.ndarray:
    """
    Adjust a covariance matrix that is non positive semidefinite
    within a given tolerance, by flipping the sign of the negative
    eigenvalues. This can occur when the covariance matrix inludes
    values that are close to zero, which results in very small
    negative numbers.

    Parameters
    ----------
    cov : `~numpy.ndarray` (D, D)
        Covariance matrix to adjust.
    tol : float, optional
        Tolerance for eigenvalues close to zero.

    Returns
    -------
    cov_psd : `~numpy.ndarray` (D, D)
        Positive semidefinite covariance matrix.
    """
    eigenvalues, eigenvectors = np.linalg.eigh(cov)
    if np.any(eigenvalues < -semidef_tol):
        raise ValueError(
            f"Covariance matrix is not positive semidefinite, {eigenvalues} above the tolerance of: {semidef_tol}"
        )
    mask = (eigenvalues < 0) & (np.abs(eigenvalues) < semidef_tol)
    eigenvalues[mask] = -eigenvalues[mask]
    cov_psd = eigenvectors @ np.diag(eigenvalues) @ eigenvectors.T
    return cov_psd


def sample_covariance_random(
    mean: np.ndarray,
    cov: np.ndarray,
    num_samples: int = 10000,
    seed: Optional[int] = None,
    semidef_tol: Optional[float] = 1e-15,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Sample a multivariate Gaussian distribution with given
    mean and covariances.

    The returned weights will be equal to 1 / num_samples so that
    each sample is equally weighted. Weights are returned from this function
    so its interface is identical to that of
    `~adam_core.coordinates.covariances.sample_covariance_sigma_points`.

    Parameters
    ----------
    mean : `~numpy.ndarray` (D)
        Multivariate mean of the Gaussian distribution.
    cov : `~numpy.ndarray` (D, D)
        Multivariate variance-covariance matrix of the Gaussian distribution.
    num_samples : int, optional
        Number of samples to draw from the distribution.
    seed : int, optional
        Seed for the random number generator.
    semidef_tol : float, optional
        Tolerance for eigenvalues close to zero.

    Returns
    -------
    samples : `~numpy.ndarray` (num_samples, D)
        The drawn samples row-by-row.
    W: `~numpy.ndarray` (num_samples)
        Weights of the samples.
    W_cov: `~numpy.ndarray` (num_samples)
        Weights of the samples to reconstruct covariance matrix.
    Raises
    ------
    ValueError : If the covariance matrix is not positive semidefinite, within the given tolerance.
    """
    # Check if the covariance matrix is non positive semidefinite. This is usually a sign
    # that something has gone wrong with the covariance. However, when the negative eigenvalues
    # are very close to zero, they can be flipped to positive without an issue. This is due to
    # the way the covariance matrix is calculated.
    if np.any(np.linalg.eigvals(cov) < 0):
        if np.any(np.linalg.eigvals(cov) < -1 * semidef_tol):
            raise ValueError(
                f"Covariance matrix is not positive semidefinite, below the tolerance of: {semidef_tol}"
            )
        else:
            logger.warning(
                "Covariance matrix is not positive semidefinite, but within tolerance, adjusting..."
            )
            cov = make_positive_semidefinite(cov)
    normal = multivariate_normal(mean=mean, cov=cov, allow_singular=True, seed=seed)
    samples = normal.rvs(num_samples)
    W = np.full(num_samples, 1 / num_samples)
    W_cov = np.full(num_samples, 1 / num_samples)
    return samples, W, W_cov


def sample_covariance_sigma_points(
    mean: np.ndarray,
    cov: np.ndarray,
    alpha: float = 1,
    beta: float = 0.0,
    kappa: float = 0.0,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Create sigma-point samples of a multivariate Gaussian distribution
    with given mean and covariances.

    Parameters
    ----------
    mean : `~numpy.ndarray` (D)
        Multivariate mean of the Gaussian distribution.
    cov : `~numpy.ndarray` (D, D)
        Multivariate variance-covariance matrix of the Gaussian distribution.
    alpha : float, optional
        Spread of the sigma points between 1e^-2 and 1.
    beta : float, optional
        Prior knowledge of the distribution usually set to 2 for a Gaussian.
    kappa : float, optional
        Secondary scaling parameter usually set to 0.

    Returns
    -------
    sigma_points : `~numpy.ndarray` (2 * D + 1, D)
        Sigma points drawn from the distribution.
    W: `~numpy.ndarray` (2 * D + 1)
        Weights of the sigma points.
    W_cov: `~numpy.ndarray` (2 * D + 1)
        Weights of the sigma points to reconstruct covariance matrix.

    References
    ----------
    [1] Wan, E. A; Van Der Merwe, R. (2000). The unscented Kalman filter for nonlinear estimation.
        Proceedings of the IEEE 2000 Adaptive Systems for Signal Processing,
        Communications, and Control Symposium, 153-158.
        https://doi.org/10.1109/ASSPCC.2000.882463
    """
    # Calculate the dimensionality of the distribution
    D = mean.shape[0]

    # See equation 15 in Wan & Van Der Merwe (2000) [1]
    N = 2 * D + 1
    sigma_points = np.empty((N, D))
    W = np.empty(N)
    W_cov = np.empty(N)

    # Calculate the scaling parameter lambda
    lambd = alpha**2 * (D + kappa) - D

    # First sigma point is the mean
    sigma_points[0] = mean

    # Beta is used to encode prior knowledge about the distribution.
    # If the distribution is a well-constrained Gaussian, beta = 2 is optimal
    # but lets set beta to 0 for now which has the effect of not weighting the mean state
    # with 0 for the covariance matrix. This is generally better for more distributions.
    # Calculate the weights for mean and the covariance matrix
    # Weight are used to reconstruct the mean and covariance matrix from the sigma points
    W[0] = lambd / (D + lambd)
    W_cov[0] = W[0] + (1 - alpha**2 + beta)

    # Take the matrix square root of the scaled covariance matrix.
    # Sometimes you'll see this done with a Cholesky decomposition for speed
    # but sqrtm is sufficiently optimized for this use case and typically provides
    # better results
    L = sqrtm((D + lambd) * cov)

    # Calculate the remaining sigma points
    for i in range(D):
        offset = L[i]
        sigma_points[i + 1] = mean + offset
        sigma_points[i + 1 + D] = mean - offset

    # The weights for the remaining sigma points are the same
    # for the mean and the covariance matrix
    W[1:] = 1 / (2 * (D + lambd))
    W_cov[1:] = 1 / (2 * (D + lambd))

    return sigma_points, W, W_cov


def weighted_mean(samples: np.ndarray, W: np.ndarray) -> np.ndarray:
    """
    Calculate the weighted mean of a set of samples.

    Parameters
    ----------
    samples : `~numpy.ndarray` (N, D)
        Samples drawn from the distribution (these can be randomly drawn
        or sigma points)
    W: `~numpy.ndarray` (N)
        Weights of the samples.

    Returns
    -------
    mean : `~numpy.ndarray` (D)
        Mean calculated from the samples and weights.
    """
    return np.dot(W, samples)


def weighted_covariance(
    mean: np.ndarray, samples: np.ndarray, W_cov: np.ndarray
) -> np.ndarray:
    """
    Calculate a covariance matrix from samples and their corresponding weights.

    Parameters
    ----------
    mean : `~numpy.ndarray` (D)
        Mean calculated from the samples and weights.
        See `~adam_core.coordinates.covariances.weighted_mean`.
    samples : `~numpy.ndarray` (N, D)
        Samples drawn from the distribution (these can be randomly drawn
        or sigma points)
    W_cov: `~numpy.ndarray` (N)
        Weights of the samples to reconstruct covariance matrix.

    Returns
    -------
    cov : `~numpy.ndarray` (D, D)
        Covariance matrix calculated from the samples and weights.
    """
    # Calculate the covariance matrix from the sigma points and weights
    # `~numpy.cov` does not support negative weights so we will calculate
    # the covariance manually
    # cov = np.cov(samples, aweights=W_cov, rowvar=False, bias=True)
    residual = samples - mean
    cov = (W_cov * residual.T) @ residual
    return cov


def transform_covariances_sampling(
    coords: np.ndarray,
    covariances: np.ndarray,
    func: Callable,
    num_samples: int = 100000,
) -> np.ma.masked_array:
    """
    Transform covariance matrices by sampling the transformation function.

    Parameters
    ----------
    coords : `~numpy.ndarray` (N, D)
        Coordinates that correspond to the input covariance matrices.
    covariances : `~numpy.ndarray` (N, D, D)
        Covariance matrices to transform via sampling.
    func : function
        A function that takes coords (N, D) as input and returns the transformed
        coordinates (N, D). See for example: `thor.coordinates.cartesian_to_spherical`
        or `thor.coordinates.cartesian_to_keplerian`.
    num_samples : int, optional
        The number of samples to draw.

    Returns
    -------
    covariances_out : `~numpy.ndarray` (N, D, D)
        Transformed covariance matrices.
    """
    covariances_out = []
    for coord, covariance in zip(coords, covariances):
        samples, W, W_cov = sample_covariance_random(coord, covariance, num_samples)
        samples_converted = func(samples)
        covariances_out.append(np.cov(samples_converted.T))

    covariances_out = np.stack(covariances_out)
    return covariances_out


def transform_covariances_jacobian(
    coords: np.ndarray,
    covariances: np.ndarray,
    _func: Callable,
    **kwargs,
) -> np.ndarray:
    """
    Transform covariance matrices by calculating the Jacobian of the transformation function
    using `~jax.jacfwd`.

    Parameters
    ----------
    coords : `~numpy.ndarray` (N, D)
        Coordinates that correspond to the input covariance matrices.
    covariances : `~numpy.ndarray` (N, D, D)
        Covariance matrices to transform via numerical differentiation.
    _func : function
        A function that takes a single coord (D) as input and return the transformed
        coordinate (D). See for example: `thor.coordinates._cartesian_to_spherical`
        or `thor.coordinates._cartesian_to_keplerian`.

    Returns
    -------
    covariances_out : `~numpy.ndarray` (N, D, D)
        Transformed covariance matrices.
    """
    jacobian = calc_jacobian(coords, _func, **kwargs)
    covariances = jacobian @ covariances @ np.transpose(jacobian, axes=(0, 2, 1))
    return covariances


def _upper_triangular_to_full(
    upper_triangular: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Convert an upper triangular matrix containing 21 elements to a full 6x6 matrix.
    """
    assert (
        len(upper_triangular) == 21
    ), "Upper triangular matrix must be a 21 element vector"

    full = np.zeros((6, 6))
    full[np.triu_indices(6)] = upper_triangular
    full[np.tril_indices(6, -1)] = full.T[np.tril_indices(6, -1)]
    return full


def _lower_triangular_to_full(
    lower_triangular: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Convert a lower triangular matrix containing 21 elements to a full 6x6 matrix.
    """
    assert (
        len(lower_triangular) == 21
    ), "Lower triangular matrix must be a 21 element vector"

    full = np.zeros((6, 6))
    full[np.tril_indices(6)] = lower_triangular
    full[np.triu_indices(6, -1)] = full.T[np.triu_indices(6, -1)]
    return full



"""
Tests for unit conversion utilities.
"""

import numpy as np

from adam_core.constants import KM_P_AU, S_P_DAY
from adam_core.coordinates.units import (
    au_per_day_to_km_per_s,
    au_to_km,
    convert_cartesian_covariance_au_to_km,
    convert_cartesian_covariance_km_to_au,
    convert_cartesian_values_au_to_km,
    convert_cartesian_values_km_to_au,
    km_per_s_to_au_per_day,
    km_to_au,
)


class TestBasicUnitConversions:
    """Test basic scalar and array unit conversion functions."""

    def test_au_to_km_scalar(self):
        """Test AU to km conversion with scalar input."""
        result = au_to_km(1.0)
        expected = KM_P_AU
        assert np.isclose(result, expected)

    def test_au_to_km_array(self):
        """Test AU to km conversion with array input."""
        input_au = np.array([1.0, 2.0, 0.5])
        result = au_to_km(input_au)
        expected = input_au * KM_P_AU
        np.testing.assert_allclose(result, expected)

    def test_km_to_au_scalar(self):
        """Test km to AU conversion with scalar input."""
        result = km_to_au(KM_P_AU)
        expected = 1.0
        assert np.isclose(result, expected)

    def test_km_to_au_array(self):
        """Test km to AU conversion with array input."""
        input_km = np.array([KM_P_AU, 2 * KM_P_AU, 0.5 * KM_P_AU])
        result = km_to_au(input_km)
        expected = np.array([1.0, 2.0, 0.5])
        np.testing.assert_allclose(result, expected)

    def test_au_per_day_to_km_per_s_scalar(self):
        """Test AU/day to km/s conversion with scalar input."""
        result = au_per_day_to_km_per_s(1.0)
        expected = KM_P_AU / S_P_DAY
        assert np.isclose(result, expected)

    def test_au_per_day_to_km_per_s_array(self):
        """Test AU/day to km/s conversion with array input."""
        input_au_day = np.array([1.0, 2.0, 0.5])
        result = au_per_day_to_km_per_s(input_au_day)
        expected = input_au_day * KM_P_AU / S_P_DAY
        np.testing.assert_allclose(result, expected)

    def test_km_per_s_to_au_per_day_scalar(self):
        """Test km/s to AU/day conversion with scalar input."""
        km_s_value = KM_P_AU / S_P_DAY
        result = km_per_s_to_au_per_day(km_s_value)
        expected = 1.0
        assert np.isclose(result, expected)

    def test_km_per_s_to_au_per_day_array(self):
        """Test km/s to AU/day conversion with array input."""
        input_km_s = np.array([KM_P_AU / S_P_DAY, 2 * KM_P_AU / S_P_DAY])
        result = km_per_s_to_au_per_day(input_km_s)
        expected = np.array([1.0, 2.0])
        np.testing.assert_allclose(result, expected)

    def test_round_trip_position_conversion(self):
        """Test that AU->km->AU gives back original values."""
        original_au = np.array([1.0, 2.5, -0.8, 10.0])
        km_values = au_to_km(original_au)
        round_trip_au = km_to_au(km_values)
        np.testing.assert_allclose(round_trip_au, original_au, rtol=1e-15)

    def test_round_trip_velocity_conversion(self):
        """Test that AU/day->km/s->AU/day gives back original values."""
        original_au_day = np.array([0.1, -0.5, 2.0, 0.01])
        km_s_values = au_per_day_to_km_per_s(original_au_day)
        round_trip_au_day = km_per_s_to_au_per_day(km_s_values)
        np.testing.assert_allclose(round_trip_au_day, original_au_day, rtol=1e-15)


class TestCartesianValuesConversion:
    """Test conversion of full 6D Cartesian coordinate arrays."""

    def test_convert_cartesian_values_au_to_km(self):
        """Test conversion of Cartesian values from AU to km."""
        # Create test data: [x, y, z, vx, vy, vz] in AU and AU/day
        values_au = np.array(
            [
                [1.0, 0.0, 0.0, 0.0, 0.01, 0.0],  # Simple case
                [-2.5, 1.5, -0.5, 0.005, -0.002, 0.001],  # More complex case
            ]
        )

        result = convert_cartesian_values_au_to_km(values_au)

        # Check positions (first 3 columns)
        expected_positions = values_au[:, :3] * KM_P_AU
        np.testing.assert_allclose(result[:, :3], expected_positions)

        # Check velocities (last 3 columns)
        expected_velocities = values_au[:, 3:] * KM_P_AU / S_P_DAY
        np.testing.assert_allclose(result[:, 3:], expected_velocities)

    def test_convert_cartesian_values_km_to_au(self):
        """Test conversion of Cartesian values from km to AU."""
        # Create test data: [x, y, z, vx, vy, vz] in km and km/s
        values_km = np.array(
            [
                [KM_P_AU, 0.0, 0.0, 0.0, KM_P_AU / S_P_DAY, 0.0],  # 1 AU, 1 AU/day
                [
                    2 * KM_P_AU,
                    -0.5 * KM_P_AU,
                    KM_P_AU,
                    0.1 * KM_P_AU / S_P_DAY,
                    0.0,
                    -0.05 * KM_P_AU / S_P_DAY,
                ],
            ]
        )

        result = convert_cartesian_values_km_to_au(values_km)

        # Check positions (first 3 columns)
        expected_positions = values_km[:, :3] / KM_P_AU
        np.testing.assert_allclose(result[:, :3], expected_positions)

        # Check velocities (last 3 columns)
        expected_velocities = values_km[:, 3:] / KM_P_AU * S_P_DAY
        np.testing.assert_allclose(result[:, 3:], expected_velocities)

    def test_round_trip_cartesian_values_conversion(self):
        """Test that AU->km->AU conversion preserves original values."""
        original_au = np.array(
            [
                [1.0, 0.0, 0.0, 0.0, 0.01, 0.0],
                [-2.5, 1.5, -0.5, 0.005, -0.002, 0.001],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # Test zeros
            ]
        )

        km_values = convert_cartesian_values_au_to_km(original_au)
        round_trip_au = convert_cartesian_values_km_to_au(km_values)

        np.testing.assert_allclose(round_trip_au, original_au, rtol=1e-15)


class TestCovarianceConversion:
    """Test covariance matrix conversion functions."""

    def test_diagonal_covariance_au_to_km(self):
        """Test conversion of diagonal covariance matrix from AU to km."""
        # Create a simple diagonal covariance matrix in AU units
        N = 2
        cov_au = np.zeros((N, 6, 6))

        # Set diagonal elements: position uncertainties in AU, velocity in AU/day
        pos_sigma_au = [0.1, 0.2, 0.05]  # AU
        vel_sigma_au_day = [0.001, 0.002, 0.0005]  # AU/day

        for i in range(N):
            for j in range(3):
                cov_au[i, j, j] = pos_sigma_au[j] ** 2  # AU²
                cov_au[i, j + 3, j + 3] = vel_sigma_au_day[j] ** 2  # (AU/day)²

        result = convert_cartesian_covariance_au_to_km(cov_au)

        # Check position variances (first 3x3 block)
        for i in range(N):
            for j in range(3):
                expected_pos_var = (pos_sigma_au[j] * KM_P_AU) ** 2
                assert np.isclose(result[i, j, j], expected_pos_var)

                expected_vel_var = (vel_sigma_au_day[j] * KM_P_AU / S_P_DAY) ** 2
                assert np.isclose(result[i, j + 3, j + 3], expected_vel_var)

    def test_diagonal_covariance_km_to_au(self):
        """Test conversion of diagonal covariance matrix from km to AU."""
        # Create a simple diagonal covariance matrix in km units
        N = 1
        cov_km = np.zeros((N, 6, 6))

        # Set diagonal elements: position uncertainties in km, velocity in km/s
        pos_sigma_km = [100.0, 200.0, 50.0]  # km
        vel_sigma_km_s = [0.1, 0.2, 0.05]  # km/s

        for j in range(3):
            cov_km[0, j, j] = pos_sigma_km[j] ** 2  # km²
            cov_km[0, j + 3, j + 3] = vel_sigma_km_s[j] ** 2  # (km/s)²

        result = convert_cartesian_covariance_km_to_au(cov_km)

        # Check position variances
        for j in range(3):
            expected_pos_var = (pos_sigma_km[j] / KM_P_AU) ** 2
            assert np.isclose(result[0, j, j], expected_pos_var)

            expected_vel_var = (vel_sigma_km_s[j] / KM_P_AU * S_P_DAY) ** 2
            assert np.isclose(result[0, j + 3, j + 3], expected_vel_var)

    def test_full_covariance_conversion(self):
        """Test conversion of full covariance matrix with off-diagonal terms."""
        # Create a covariance matrix with correlations between position and velocity
        N = 1
        cov_au = np.zeros((N, 6, 6))

        # Set some diagonal terms
        cov_au[0, 0, 0] = 0.01  # x variance (AU²)
        cov_au[0, 1, 1] = 0.02  # y variance (AU²)
        cov_au[0, 3, 3] = 0.0001  # vx variance ((AU/day)²)
        cov_au[0, 4, 4] = 0.0002  # vy variance ((AU/day)²)

        # Set some off-diagonal terms (correlations)
        cov_au[0, 0, 1] = cov_au[0, 1, 0] = 0.005  # x-y correlation (AU²)
        cov_au[0, 0, 3] = cov_au[0, 3, 0] = 0.001  # x-vx correlation (AU·AU/day)
        cov_au[0, 1, 4] = cov_au[0, 4, 1] = 0.002  # y-vy correlation (AU·AU/day)
        cov_au[0, 3, 4] = cov_au[0, 4, 3] = 0.00005  # vx-vy correlation ((AU/day)²)

        result_km = convert_cartesian_covariance_au_to_km(cov_au)

        # Test diagonal elements
        assert np.isclose(result_km[0, 0, 0], 0.01 * KM_P_AU**2)  # x variance
        assert np.isclose(result_km[0, 1, 1], 0.02 * KM_P_AU**2)  # y variance
        assert np.isclose(
            result_km[0, 3, 3], 0.0001 * (KM_P_AU / S_P_DAY) ** 2
        )  # vx variance
        assert np.isclose(
            result_km[0, 4, 4], 0.0002 * (KM_P_AU / S_P_DAY) ** 2
        )  # vy variance

        # Test off-diagonal elements
        assert np.isclose(result_km[0, 0, 1], 0.005 * KM_P_AU**2)  # x-y correlation
        assert np.isclose(result_km[0, 1, 0], 0.005 * KM_P_AU**2)  # symmetry
        assert np.isclose(
            result_km[0, 0, 3], 0.001 * KM_P_AU * KM_P_AU / S_P_DAY
        )  # x-vx correlation
        assert np.isclose(
            result_km[0, 3, 0], 0.001 * KM_P_AU * KM_P_AU / S_P_DAY
        )  # symmetry

    def test_round_trip_covariance_conversion(self):
        """Test that AU->km->AU covariance conversion preserves original values."""
        # Create a random symmetric positive definite covariance matrix
        N = 3
        np.random.seed(42)  # For reproducibility

        # Generate random covariance matrices
        cov_au_original = np.zeros((N, 6, 6))
        for i in range(N):
            # Create a random positive definite matrix
            A = np.random.randn(6, 6)
            cov_au_original[i] = A @ A.T * 0.01  # Scale to reasonable values

        # Convert AU -> km -> AU
        cov_km = convert_cartesian_covariance_au_to_km(cov_au_original)
        cov_au_round_trip = convert_cartesian_covariance_km_to_au(cov_km)

        np.testing.assert_allclose(cov_au_round_trip, cov_au_original, rtol=1e-13)

    def test_covariance_conversion_preserves_symmetry(self):
        """Test that covariance conversion preserves matrix symmetry."""
        # Create a symmetric covariance matrix
        N = 2
        cov_au = np.random.randn(N, 6, 6) * 0.01

        # Make it symmetric
        for i in range(N):
            cov_au[i] = (cov_au[i] + cov_au[i].T) / 2

        result_km = convert_cartesian_covariance_au_to_km(cov_au)

        # Check that result is still symmetric
        for i in range(N):
            np.testing.assert_allclose(result_km[i], result_km[i].T, rtol=1e-15)

    def test_covariance_conversion_known_values(self):
        """Test covariance conversion with known manual calculation."""
        # Test a simple case where we can manually verify the calculation
        N = 1
        cov_au = np.zeros((N, 6, 6))

        # Set one position-position term and one velocity-velocity term
        cov_au[0, 0, 0] = 1.0  # 1 AU²
        cov_au[0, 3, 3] = 1.0  # 1 (AU/day)²

        result = convert_cartesian_covariance_au_to_km(cov_au)

        # Manual calculation:
        # Position variance: 1 AU² × (KM_P_AU)² = KM_P_AU² km²
        # Velocity variance: 1 (AU/day)² × (KM_P_AU/S_P_DAY)² = (KM_P_AU/S_P_DAY)² (km/s)²

        expected_pos_var = KM_P_AU**2
        expected_vel_var = (KM_P_AU / S_P_DAY) ** 2

        assert np.isclose(result[0, 0, 0], expected_pos_var)
        assert np.isclose(result[0, 3, 3], expected_vel_var)


class TestEdgeCases:
    """Test edge cases and error conditions."""

    def test_zero_values(self):
        """Test that zero values are handled correctly."""
        assert au_to_km(0.0) == 0.0
        assert km_to_au(0.0) == 0.0
        assert au_per_day_to_km_per_s(0.0) == 0.0
        assert km_per_s_to_au_per_day(0.0) == 0.0

    def test_negative_values(self):
        """Test that negative values are handled correctly."""
        assert au_to_km(-1.0) == -KM_P_AU
        assert km_to_au(-KM_P_AU) == -1.0

    def test_very_small_values(self):
        """Test conversion of very small values."""
        tiny_au = 1e-10
        result_km = au_to_km(tiny_au)
        round_trip = km_to_au(result_km)
        assert np.isclose(round_trip, tiny_au, rtol=1e-15)

    def test_very_large_values(self):
        """Test conversion of very large values."""
        large_au = 1e6
        result_km = au_to_km(large_au)
        round_trip = km_to_au(result_km)
        assert np.isclose(round_trip, large_au, rtol=1e-15)



"""
This module tests using custom SPICE kernels using a small JWST kernel.
"""

from pathlib import Path

import numpy as np
import pytest
import spiceypy as sp

from ...coordinates.origin import OriginCodes
from ...time import Timestamp
from ...utils.spice import (
    get_spice_body_state,
    register_spice_kernel,
    unregister_spice_kernel,
)
from ..observers import Observers
from ..state import get_observer_state

# JWST SPICE kernel paths
JWST_KERNEL_DIR = (
    Path(__file__).parent.parent.parent / "utils" / "tests" / "data" / "spice"
)
JWST_KERNEL_PATH = JWST_KERNEL_DIR / "jwst_horizons_20200101_20240101_v01.bsp"


@pytest.fixture
def jwst_kernel():
    """Fixture to handle JWST kernel registration and cleanup."""
    if not JWST_KERNEL_PATH.exists():
        pytest.skip(
            "JWST SPICE kernel not found. Please download it to tests/data/spice/jwst_horizons_20200101_20240101_v01.bsp"
        )

    register_spice_kernel(str(JWST_KERNEL_PATH))
    yield
    unregister_spice_kernel(str(JWST_KERNEL_PATH))


def test_jwst_as_observer(jwst_kernel):
    """Test that we can use JWST as an observer using string identifier."""
    # Test with a date known to be in the JWST kernel range
    test_time = Timestamp.from_iso8601(["2022-01-01T00:00:00Z"])

    # Get JWST state vectors using the observer system with string identifier
    jwst_observer_state = get_observer_state(
        "JWST", test_time, frame="ecliptic", origin=OriginCodes.SUN
    )

    # Verify we got valid state vectors
    assert jwst_observer_state.values.shape == (1, 6)
    assert jwst_observer_state.frame == "ecliptic"
    assert jwst_observer_state.origin.code[0].as_py() == "SUN"

    # Get JWST state using the direct spice function
    jwst_id = sp.bodn2c("JWST")
    jwst_direct_state = get_spice_body_state(
        jwst_id, test_time, frame="ecliptic", origin=OriginCodes.SUN
    )

    # Both methods should return identical results
    np.testing.assert_allclose(
        jwst_observer_state.values, jwst_direct_state.values, rtol=1e-10
    )


def test_jwst_as_observer_from_code(jwst_kernel):
    """Test that we can use JWST as an observer using Observers.from_code."""
    # Test with a date known to be in the JWST kernel range
    test_time = Timestamp.from_iso8601(["2022-01-01T00:00:00Z"])

    observers = Observers.from_code("JWST", test_time)
    assert len(observers) == 1
    assert observers.code[0].as_py() == "JWST"
    assert observers.coordinates.values.shape == (1, 6)
    assert observers.coordinates.frame == "ecliptic"
    assert observers.coordinates.origin.code[0].as_py() == "SUN"

    observers = Observers.from_codes(["JWST"], test_time)
    assert len(observers) == 1
    assert observers.code[0].as_py() == "JWST"
    assert observers.coordinates.values.shape == (1, 6)
    assert observers.coordinates.frame == "ecliptic"
    assert observers.coordinates.origin.code[0].as_py() == "SUN"



import numpy as np
import pyarrow.compute as pc
import pytest
from adam_assist import ASSISTPropagator

from ...coordinates import CoordinateCovariances, SphericalCoordinates
from ...coordinates.origin import Origin
from ...observers import Observers
from ...utils.helpers.observations import make_observations
from ...utils.helpers.orbits import make_real_orbits
from ..evaluate import OrbitDeterminationObservations
from ..fitted_orbits import FittedOrbitMembers, FittedOrbits
from ..od import od


@pytest.fixture
def real_data():
    # Generate real observations and orbits
    exposures, detections, associations = make_observations()
    orbits = make_real_orbits(num_orbits=18)

    # Select Ivezic for testing
    object_id = "202930 Ivezic (1998 SG172)"
    orbit = orbits.select("object_id", object_id)

    # Filter observations for the selected object ID
    associations_i = associations.select("object_id", object_id)
    detections_i = detections.apply_mask(
        pc.is_in(detections.id, associations_i.detection_id)
    )

    exposures_i = exposures.apply_mask(pc.is_in(exposures.id, detections_i.exposure_id))

    sigmas = np.full((len(detections_i.ra_sigma), 6), np.nan)
    sigmas[:, 1] = detections_i.ra_sigma.to_numpy(zero_copy_only=False)
    sigmas[:, 2] = detections_i.dec_sigma.to_numpy(zero_copy_only=False)
    coordinates = SphericalCoordinates.from_kwargs(
        lon=detections_i.ra.to_numpy(),
        lat=detections_i.dec.to_numpy(),
        covariance=CoordinateCovariances.from_sigmas(sigmas),
        time=detections_i.time,
        origin=Origin.from_kwargs(code=exposures_i.observatory_code),
        frame="equatorial",  # Assuming the frame is equatorial
    )

    # Generate Observers from exposures start_time and observatory codes
    observers = Observers.from_codes(
        times=detections_i.time, codes=exposures_i.observatory_code
    )

    observations = OrbitDeterminationObservations.from_kwargs(
        id=detections_i.id.to_numpy(zero_copy_only=False),
        coordinates=coordinates,
        observers=observers,
    )

    # Use the first orbit as the starting orbit
    starting_orbit = orbit

    return starting_orbit, observations


def test_od(real_data):
    starting_orbit, observations = real_data

    # Run the orbit determination
    od_orbit, od_orbit_members = od(
        orbit=starting_orbit,
        observations=observations[:10],
        propagator=ASSISTPropagator,
        rchi2_threshold=10,
        min_obs=3,
        min_arc_length=1.0,
        contamination_percentage=0.0,
        delta=1e-6,
        max_iter=20,
        method="central",
        propagator_kwargs={},
    )
    # Assertions

    assert isinstance(od_orbit, FittedOrbits)
    assert isinstance(od_orbit_members, FittedOrbitMembers)
    assert len(od_orbit) == 1
    assert len(od_orbit_members) == 10
    assert od_orbit.success[0]



from typing import Tuple

import numpy as np
import pyarrow.compute as pc

from .differential_correction import OrbitDeterminationObservations
from .fitted_orbits import FittedOrbitMembers


def calculate_max_outliers(
    num_obs: int, min_obs: int, contamination_percentage: float
) -> int:
    """
    Calculate the maximum number of allowable outliers. Linkages may contain err
    oneuos observations that need to be removed. This function calculates the maximum number of
    observations that can be removed before the linkage no longer has the minimum number
    of observations required. The contamination percentage is the maximum percentage of observations
    that allowed to be erroneous.

    Parameters
    ----------
    num_obs : int
        Number of observations in the linkage.
    min_obs : int
        Minimum number of observations required for a valid linkage.
    contamination_percentage : float
        Maximum percentage of observations that allowed to be erroneous. Range is [0, 100].

    Returns
    -------
    outliers : int
        Maximum number of allowable outliers.
    """
    assert (
        num_obs >= min_obs
    ), "Number of observations must be greater than or equal to the minimum number of observations."
    assert (
        contamination_percentage >= 0 and contamination_percentage <= 100
    ), "Contamination percentage must be between 0 and 100."

    max_outliers = num_obs * (contamination_percentage / 100)
    outliers = np.min([max_outliers, num_obs - min_obs]).astype(int)
    return outliers


def remove_lowest_probability_observation(
    orbit_members: FittedOrbitMembers, observations: OrbitDeterminationObservations
) -> Tuple[str, OrbitDeterminationObservations]:
    """
    Remove the observation with the worst residual from the observations. The probability is defined to be
    the probability of drawing a more extreme residual than the one observed. If multiple observations
    have the same probability, then the observation with the highest squared residual value is removed.

    Parameters
    ----------
    orbit_members : FittedOrbitMembers
        The orbit members that contain the residuals with respect to the observations.
    observations : OrbitDeterminationObservations
        The observations to remove the worst residual from.

    Returns
    -------
    obs_id : str
        The ID of the observation that was removed.
    filtered_observations : OrbitDeterminationObservations
        The observations with the worst residual removed.
    """
    assert (
        len(pc.unique(orbit_members.orbit_id)) == 1
    ), "Orbit members must only contain one orbit"
    assert pc.all(
        pc.is_in(orbit_members.obs_id, observations.id)
    ).as_py(), "Observations must contain all orbit member observations"

    # Find the worst outlier (the observation that has the lowest probability of
    # drawing a more extreme residual than the one observed)
    worst_outlier = orbit_members.apply_mask(
        pc.equal(
            orbit_members.residuals.probability,
            pc.min(orbit_members.residuals.probability),
        )
    )

    if len(worst_outlier) > 1:
        # If there are multiple worst outliers (which would be quite unlikely),
        # then remove the outlier with the highest squared residual value
        index = np.nansum(worst_outlier.residuals.to_array() ** 2, axis=1).argmax()
        worst_outlier = worst_outlier.take([index])

    # Grab the observation ID of the worst outlier
    obs_id = worst_outlier.obs_id[0].as_py()

    # Grab the surviving observation IDs
    obs_ids = orbit_members.apply_mask(
        pc.invert(pc.equal(orbit_members.obs_id, obs_id))
    ).obs_id

    return obs_id, observations.apply_mask(pc.is_in(observations.id, obs_ids))



# flake8: noqa: F401

# TODO: associations - should it be in the API? its useful for test helpers, but kind of niche
from .ades import (
    ADES_to_string,
    ADESObservations,
    ObsContext,
    ObservatoryObsContext,
    SoftwareObsContext,
    SubmitterObsContext,
    TelescopeObsContext,
)
from .associations import Associations
from .detections import PointSourceDetections
from .exposures import Exposures
from .photometry import Photometry
from .source_catalog import SourceCatalog

__all__ = [
    "Associations",
    "Exposures",
    "PointSourceDetections",
    "Photometry",
    "SourceCatalog",
]



"""
Tests for generating ephemeris using custom SPICE kernels with JWST as an observer.
This file demonstrates how to register custom kernels and create observers from them.
"""

from pathlib import Path

import numpy as np
import pyarrow.compute as pc
import pytest
from adam_assist import ASSISTPropagator

from ...coordinates.cartesian import CartesianCoordinates
from ...coordinates.origin import Origin
from ...observers.observers import Observers
from ...orbits.orbits import Orbits
from ...time import Timestamp
from ...utils.helpers.orbits import make_real_orbits
from ...utils.spice import register_spice_kernel, unregister_spice_kernel

# JWST SPICE kernel paths
JWST_KERNEL_DIR = (
    Path(__file__).parent.parent.parent / "utils" / "tests" / "data" / "spice"
)
JWST_KERNEL_PATH = JWST_KERNEL_DIR / "jwst_horizons_20200101_20240101_v01.bsp"


@pytest.fixture
def jwst_kernel():
    """Fixture to handle JWST kernel registration and cleanup."""
    if not JWST_KERNEL_PATH.exists():
        pytest.skip(
            "JWST SPICE kernel not found. Please download it to tests/data/spice/jwst_horizons_20200101_20240101_v01.bsp"
        )

    register_spice_kernel(str(JWST_KERNEL_PATH))
    yield
    unregister_spice_kernel(str(JWST_KERNEL_PATH))


@pytest.mark.parametrize("max_processes", [1, 4])
def test_generate_ephemeris_with_custom_kernel(jwst_kernel, max_processes):
    """Test generating ephemeris using the JWST kernel with ASSISTPropagator."""
    # Create observation times within the JWST kernel range (2020-01-01 to 2024-01-01)
    obs_times = Timestamp.from_iso8601(
        [
            "2022-06-15T00:00:00Z",
            "2022-06-15T06:00:00Z",
            "2022-06-15T12:00:00Z",
            "2022-06-15T18:00:00Z",
        ]
    )

    # Create JWST observers using the registered kernel
    jwst_observers = Observers.from_code("JWST", obs_times)

    # Create some test orbits - using the helper function
    orbits = make_real_orbits(5)

    # Initialize the ASSISTPropagator
    propagator = ASSISTPropagator()

    # Generate ephemeris using the custom kernel observer
    ephemeris = propagator.generate_ephemeris(
        orbits, jwst_observers, max_processes=max_processes
    )

    # Verify basic properties of the generated ephemeris
    assert len(ephemeris) == len(orbits) * len(jwst_observers)

    # Check that observer is correctly set to JWST
    assert np.all(
        [code.as_py() == "JWST" for code in ephemeris.coordinates.origin.code]
    )

    # Check that the ephemeris time scale is UTC (as expected for output)
    assert ephemeris.coordinates.time.scale == "utc"

    # Verify light time calculations are reasonable
    # Light time should be positive and less than ~1 hour (0.04 days) for Solar System objects
    assert np.all(ephemeris.light_time.to_numpy(zero_copy_only=False) > 0)
    assert np.all(ephemeris.light_time.to_numpy(zero_copy_only=False) < 0.04)

    # Verify that coordinates.time - aberrated_coordinates.time equals light_time.
    # Rescale both to the same time scale before taking the difference.
    coords_tdb = ephemeris.coordinates.time.rescale("tdb")
    aberr_tdb = ephemeris.aberrated_coordinates.time.rescale("tdb")
    time_difference_days, time_difference_nanos = coords_tdb.difference(aberr_tdb)
    fractional_days = pc.divide(time_difference_nanos, 86400 * 1e9)
    time_difference = pc.add(time_difference_days, fractional_days)
    np.testing.assert_allclose(
        time_difference.to_numpy(zero_copy_only=False),
        ephemeris.light_time.to_numpy(zero_copy_only=False),
        atol=1e-6,
    )


@pytest.mark.parametrize("max_processes", [1, 4])
def test_mixed_observer_ephemeris(jwst_kernel, max_processes):
    """Test generating ephemeris with both JWST and ground-based observers."""

    # Create combined observer list with both JWST and a ground station
    all_observers = Observers.from_codes(
        ["JWST", "500", "JWST", "500"],  # JWST and geocenter (500)
        Timestamp.from_iso8601(
            [
                "2022-06-15T00:00:00Z",
                "2022-06-15T00:00:00Z",
                "2022-06-15T12:00:00Z",
                "2022-06-15T12:00:00Z",
            ]
        ),
    )

    # Create specific test orbits with known properties
    test_orbits = Orbits.from_kwargs(
        orbit_id=["near_earth", "distant"],
        object_id=["near_earth", "distant"],
        coordinates=CartesianCoordinates.from_kwargs(
            # First orbit close to Earth, second one further away
            x=[1.0, 3.0],
            y=[0.0, 0.0],
            z=[0.0, 0.0],
            vx=[0.0, 0.0],
            vy=[1.0, 0.6],  # Different orbital velocities
            vz=[0.0, 0.0],
            # Need to pick a time close to JWST spice kernel observer times so we don't
            # launch out of the solar system
            time=Timestamp.from_iso8601(
                ["2022-06-14T00:00:00Z", "2022-06-14T00:00:00Z"]
            ),  # Same start time for both orbits
            origin=Origin.from_kwargs(code=["SUN", "SUN"]),
            frame="ecliptic",
        ),
    )

    # Generate ephemeris with mixed observer types
    propagator = ASSISTPropagator()
    ephemeris = propagator.generate_ephemeris(
        test_orbits, all_observers, max_processes=max_processes
    )

    # Verify we got ephemeris for all combinations
    assert len(ephemeris) == len(test_orbits) * len(all_observers)

    # Check that we have both types of observers in the results
    observer_codes = [code.as_py() for code in ephemeris.coordinates.origin.code]
    assert "JWST" in observer_codes
    assert "500" in observer_codes

    # Separate ephemeris by observer type
    jwst_ephemeris = ephemeris.select("coordinates.origin.code", "JWST")
    geo_ephemeris = ephemeris.select("coordinates.origin.code", "500")

    # Both should have the same number of rows
    assert len(jwst_ephemeris) == len(geo_ephemeris)

    # For the near-Earth object, there should be a noticeable difference in position
    # as observed from JWST vs. Earth for the same timestamp
    near_earth_jwst = jwst_ephemeris.select("orbit_id", "near_earth")
    near_earth_geo = geo_ephemeris.select("orbit_id", "near_earth")

    # Extract spherical coordinates to compare the angular differences
    jwst_ra = near_earth_jwst.coordinates.lon.to_numpy(zero_copy_only=False)
    geo_ra = near_earth_geo.coordinates.lon.to_numpy(zero_copy_only=False)
    jwst_dec = near_earth_jwst.coordinates.lat.to_numpy(zero_copy_only=False)
    geo_dec = near_earth_geo.coordinates.lat.to_numpy(zero_copy_only=False)

    # There should be measurable parallax between Earth and JWST for near objects
    # (at least a few arcseconds)
    ra_diff = np.abs(jwst_ra - geo_ra)
    dec_diff = np.abs(jwst_dec - geo_dec)

    # Check that there's at least some difference in the observed positions
    assert np.any(ra_diff > 1e-3) or np.any(dec_diff > 1e-3)

    # Verify linkage to observers works
    linkage = ephemeris.link_to_observers(all_observers)
    assert len(linkage.all_unique_values) == len(all_observers)



import multiprocessing as mp
from typing import Dict, List, Optional, Tuple

import jax.numpy as jnp
import numpy as np
import quivr as qv
import ray
from jax import config, jit, vmap
from ray import ObjectRef

from ..coordinates.cartesian import CartesianCoordinates
from ..coordinates.covariances import (
    CoordinateCovariances,
    transform_covariances_jacobian,
)
from ..coordinates.origin import Origin
from ..orbits.orbits import Orbits
from ..ray_cluster import initialize_use_ray
from ..time import Timestamp
from ..utils.chunking import process_in_chunks
from ..utils.iter import _iterate_chunks
from .lagrange import apply_lagrange_coefficients, calc_lagrange_coefficients

config.update("jax_enable_x64", True)


@jit
def _propagate_2body(
    orbit: jnp.ndarray,
    t0: float,
    t1: float,
    mu: float,
    max_iter: int = 1000,
    tol: float = 1e-14,
) -> jnp.ndarray:
    """
    Propagate an orbit from t0 to t1.

    Parameters
    ----------
    orbit : `~jax.numpy.ndarray` (6)
        Cartesian orbit with position in units of au and velocity in units of au per day.
    t0 : float (1)
        Epoch in MJD at which the orbit are defined.
    t1 : float (N)
        Epochs to which to propagate the given orbit.
    mu : float (1)
        Gravitational parameter (GM) of the attracting body in units of
        au**3 / d**2.
    max_iter : int, optional
        Maximum number of iterations over which to converge. If number of iterations is
        exceeded, will return the value of the universal anomaly at the last iteration.
    tol : float, optional
        Numerical tolerance to which to compute universal anomaly using the Newtown-Raphson
        method.

    Returns
    -------
    orbits : `~jax.numpy.ndarray` (N, 6)
        Orbit propagated to each MJD with position in units of au and velocity in units
        of au per day.
    """
    r = orbit[0:3]
    v = orbit[3:6]
    dt = t1 - t0

    lagrange_coeffs, stumpff_coeffs, chi = calc_lagrange_coefficients(
        r, v, dt, mu=mu, max_iter=max_iter, tol=tol
    )
    r_new, v_new = apply_lagrange_coefficients(r, v, *lagrange_coeffs)

    return jnp.array([r_new[0], r_new[1], r_new[2], v_new[0], v_new[1], v_new[2]])


# Vectorization Map: _propagate_2body
_propagate_2body_vmap = jit(
    vmap(_propagate_2body, in_axes=(0, 0, 0, 0, None, None), out_axes=(0))
)


def _propagate_2body_serial(
    orbits: Orbits,
    times: Timestamp,
    *,
    max_iter: int,
    tol: float,
) -> Orbits:
    """
    Serial (single-process) implementation of 2-body propagation.

    The Ray backend uses this function inside each worker.
    """
    # Extract and prepare data
    cartesian_orbits = orbits.coordinates.values
    t0 = orbits.coordinates.time.rescale("tdb").mjd()
    t1 = times.rescale("tdb").mjd()
    mu = orbits.coordinates.origin.mu()
    orbit_ids = orbits.orbit_id.to_numpy(zero_copy_only=False)
    object_ids = orbits.object_id.to_numpy(zero_copy_only=False)

    # Fixed chunk size to keep JAX shapes stable.
    chunk_size = 200

    n_orbits = cartesian_orbits.shape[0]
    n_times = len(times)
    orbit_ids_ = np.repeat(orbit_ids, n_times)
    object_ids_ = np.repeat(object_ids, n_times)
    orbits_array_ = np.repeat(cartesian_orbits, n_times, axis=0)
    mu_ = np.repeat(mu, n_times)
    t0_ = np.repeat(t0, n_times)
    t1_ = np.tile(t1, n_orbits)

    # Preserve physical parameters by repeating per-orbit rows across times.
    pp_idx = np.repeat(np.arange(n_orbits), n_times).tolist()
    physical_parameters_ = orbits.physical_parameters.take(pp_idx)

    num_entries = n_orbits * n_times
    orbits_propagated = np.empty((num_entries, 6), dtype=np.float64)
    start = 0
    for orbits_chunk, t0_chunk, t1_chunk, mu_chunk in zip(
        process_in_chunks(orbits_array_, chunk_size),
        process_in_chunks(t0_, chunk_size),
        process_in_chunks(t1_, chunk_size),
        process_in_chunks(mu_, chunk_size),
    ):
        valid = min(chunk_size, num_entries - start)
        orbits_propagated_chunk = _propagate_2body_vmap(
            orbits_chunk, t0_chunk, t1_chunk, mu_chunk, max_iter, tol
        )
        orbits_propagated[start : start + valid] = np.asarray(orbits_propagated_chunk)[
            :valid
        ]
        start += valid

    if start != num_entries:
        raise RuntimeError(
            f"Internal error: expected {num_entries} propagated rows, got {start}"
        )

    if not orbits.coordinates.covariance.is_all_nan():
        cartesian_covariances = orbits.coordinates.covariance.to_matrix()
        covariances_array_ = np.repeat(cartesian_covariances, n_times, axis=0)

        cartesian_covariances = transform_covariances_jacobian(
            orbits_array_,
            covariances_array_,
            _propagate_2body,
            in_axes=(0, 0, 0, 0, None, None),
            out_axes=0,
            t0=t0_,
            t1=t1_,
            mu=mu_,
            max_iter=max_iter,
            tol=tol,
        )
        cartesian_covariances = CoordinateCovariances.from_matrix(cartesian_covariances)
    else:
        cartesian_covariances = None

    origin_code = np.repeat(
        orbits.coordinates.origin.code.to_numpy(zero_copy_only=False), n_times
    )

    return Orbits.from_kwargs(
        orbit_id=orbit_ids_,
        object_id=object_ids_,
        physical_parameters=physical_parameters_,
        coordinates=CartesianCoordinates.from_kwargs(
            x=orbits_propagated[:, 0],
            y=orbits_propagated[:, 1],
            z=orbits_propagated[:, 2],
            vx=orbits_propagated[:, 3],
            vy=orbits_propagated[:, 4],
            vz=orbits_propagated[:, 5],
            covariance=cartesian_covariances,
            time=Timestamp.from_mjd(t1_, scale="tdb"),
            origin=Origin.from_kwargs(code=origin_code),
            frame="ecliptic",
        ),
    )


@ray.remote
def propagate_2body_worker_ray(
    start: int,
    idx_chunk: np.ndarray,
    orbits: Orbits,
    times: Timestamp,
    max_iter: int,
    tol: float,
) -> Tuple[int, Orbits]:
    orbits_chunk = orbits.take(idx_chunk)
    propagated = _propagate_2body_serial(
        orbits_chunk, times, max_iter=max_iter, tol=tol
    )
    return start, propagated


def propagate_2body(
    orbits: Orbits,
    times: Timestamp,
    max_iter: int = 1000,
    tol: float = 1e-14,
    *,
    max_processes: Optional[int] = 1,
    chunk_size: int = 100,
) -> Orbits:
    """
    Propagate orbits using the 2-body universal anomaly formalism.

    Parameters
    ----------
    orbits : `~adam_core.orbits.orbits.Orbits` (N)
        Cartesian orbits with position in units of au and velocity in units of au per day.
    times : Timestamp (M)
        Epochs to which to propagate each orbit. If a single epoch is given, all orbits are propagated to this
        epoch. If multiple epochs are given, then each orbit to will be propagated to each epoch.
    max_iter : int, optional
        Maximum number of iterations over which to converge. If number of iterations is
        exceeded, will return the value of the universal anomaly at the last iteration.
    tol : float, optional
        Numerical tolerance to which to compute universal anomaly using the Newtown-Raphson
        method.

    Returns
    -------
    orbits : `~adam_core.orbits.orbits.Orbits` (N*M)
        Orbits propagated to each MJD.
    """
    if max_processes is None:
        max_processes = mp.cpu_count()

    if max_processes <= 1:
        return _propagate_2body_serial(orbits, times, max_iter=max_iter, tol=tol)

    initialize_use_ray(num_cpus=max_processes)

    # Put large inputs in object store once.
    orbits_ref = ray.put(orbits)  # type: ignore[name-defined]
    times_ref = ray.put(times)  # type: ignore[name-defined]

    idx = np.arange(0, len(orbits), dtype=np.int64)
    pending: List["ObjectRef"] = []  # type: ignore[name-defined]
    results: Dict[int, Orbits] = {}

    for idx_chunk in _iterate_chunks(idx, chunk_size):
        start = int(idx_chunk[0]) if len(idx_chunk) else 0
        pending.append(
            propagate_2body_worker_ray.remote(  # type: ignore[name-defined]
                start, idx_chunk, orbits_ref, times_ref, max_iter, tol
            )
        )

        if len(pending) >= max_processes * 1.5:
            finished, pending = ray.wait(pending, num_returns=1)  # type: ignore[name-defined]
            start_i, propagated_i = ray.get(finished[0])  # type: ignore[name-defined]
            results[int(start_i)] = propagated_i

    while pending:
        finished, pending = ray.wait(pending, num_returns=1)  # type: ignore[name-defined]
        start_i, propagated_i = ray.get(finished[0])  # type: ignore[name-defined]
        results[int(start_i)] = propagated_i

    chunks = [results[k] for k in sorted(results.keys())]
    return qv.concatenate(chunks) if chunks else Orbits.empty()
